"""
Copyright 2024 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
import logging
from typing import Dict, List, Tuple

from gs_quant.session import GsSession
from gs_quant.target.risk import Scenario

_logger = logging.getLogger(__name__)


class GsScenarioApi:
    """GS Scenarios API client implementation"""

    @classmethod
    def create_scenario(cls, scenario: Scenario) -> Scenario:
        return GsSession.current._post('/risk/scenarios', scenario, cls=Scenario)

    @classmethod
    def get_scenario(cls, scenario_id: str) -> Scenario:
        return GsSession.current._get(f'/risk/scenarios/{scenario_id}', cls=Scenario)

    @classmethod
    def get_many_scenarios(cls,
                           ids: List[str] = None,
                           names: List[str] = None,
                           limit: int = 100,
                           **kwargs) -> Tuple[Scenario, ...]:
        url = f'/risk/scenarios?limit={limit}'
        if ids:
            url += f'&id={"&id=".join(ids)}'
        if names:
            url += f'&name={"&name=".join(names)}'
        if kwargs:
            for k, v in kwargs.items():
                url += f'&{k}={f"&{k}=".join(v)}' if isinstance(v, list) else f'&{k}={v}'

        return GsSession.current._get(url, cls=Scenario).get('results', [])

    @classmethod
    def get_scenario_by_name(cls, name: str) -> Scenario:
        url = f"/risk/scenarios?name={name}"
        ret = GsSession.current._get(url, cls=Scenario)
        num_found = ret.get('totalResults', 0)

        if num_found == 0:
            raise ValueError(f'Scenario {name}not found')
        elif num_found > 1:
            raise ValueError(f'More than one scemario named {name}')
        else:
            return ret['results'][0]

    @classmethod
    def update_scenario(cls, scenario: Scenario) -> Dict:
        return GsSession.current._put(f'/risk/scenarios/{scenario.id_}', scenario, cls=Scenario)

    @classmethod
    def delete_scenario(cls, scenario_id: str) -> Dict:
        return GsSession.current._delete(f'/risk/scenarios/{scenario_id}')

    @classmethod
    def calculate_scenario(cls, request: Dict) -> Dict:
        return GsSession.current._post('/scenarios/calculate', request)


class GsFactorScenarioApi(GsScenarioApi):
    def __init__(self):
        super().__init__()

    @classmethod
    def get_many_scenarios(cls,
                           ids: List[str] = None,
                           names: List[str] = None,
                           limit: int = 100,
                           type: str = None,
                           risk_model: str = None,
                           shocked_factors: List[str] = None,
                           shocked_factor_categories: List[str] = None,
                           start_date: dt.date = None,
                           end_date: dt.date = None,
                           tags: List[str] = None) -> Tuple[Scenario, ...]:
        factor_scenario_args = {}
        if risk_model:
            factor_scenario_args['riskModel'] = risk_model
        if type:
            factor_scenario_args['factorScenarioType'] = type
        if shocked_factors:
            factor_scenario_args['shockedFactor'] = shocked_factors
        if shocked_factor_categories:
            factor_scenario_args['shockedFactorCategory'] = shocked_factor_categories
        if start_date:
            factor_scenario_args['historicalSimulationStartDate'] = start_date
        if end_date:
            factor_scenario_args['historicalSimulationEndDate'] = end_date
        if tags:
            factor_scenario_args['tags'] = tags

        many_scenarios = super().get_many_scenarios(ids=ids, names=names, limit=limit, **factor_scenario_args)
        many_scenarios = tuple([scenario for scenario in many_scenarios if scenario.type_])

        return many_scenarios

    @classmethod
    def calculate_scenario(cls, calculation_request: Dict) -> Dict:
        return super().calculate_scenario(request=calculation_request)



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt
import pandas as pd

from typing import Dict, Any, Tuple, Union

from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.generic_datatype_encoders import \
    decode_inst_tuple, decode_inst
from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.risk_result_datatype_encoders import \
    encode_series_result, encode_dataframe_result
from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.risk_result_encoders import decode_risk_result, \
    decode_risk_result_with_data
from gs_quant.api.gs.backtests_xasset.response_datatypes.backtest_datatypes import Transaction, TransactionDirection
from gs_quant.api.gs.backtests_xasset.response_datatypes.risk_result_datatypes import RiskResultWithData
from gs_quant.common import Currency, CurrencyName, RiskMeasure
from gs_quant.json_convertors_common import encode_risk_measure, decode_risk_measure
from gs_quant.priceable import PriceableImpl
from gs_quant.target.backtests import FlowVolBacktestMeasure


def encode_response_obj(data: Any) -> Dict:
    if isinstance(data, RiskMeasure):
        return encode_risk_measure(data)
    if isinstance(data, pd.Series):
        return encode_series_result(data)
    if isinstance(data, pd.DataFrame):
        return encode_dataframe_result(data)
    return data.to_dict()


def decode_leg_refs(d: dict) -> Dict[str, PriceableImpl]:
    return {k: decode_inst(v) for k, v in d.items()}


def decode_risk_measure_refs(d: dict) -> Dict[str, RiskMeasure]:
    return {k: decode_risk_measure(v) for k, v in d.items()}


def decode_result_tuple(results: tuple):
    return tuple(decode_risk_result(r) for r in results)


def decode_basic_bt_measure_dict(results: dict) -> Dict[FlowVolBacktestMeasure, Dict[dt.date, RiskResultWithData]]:
    return {FlowVolBacktestMeasure(k): {dt.date.fromisoformat(d): decode_risk_result_with_data(r) for d, r in v.items()}
            for k, v in results.items()}


def decode_basic_bt_transactions(results: dict, decode_instruments: bool = True) -> \
        Dict[dt.date, Tuple[Transaction, ...]]:
    def to_ccy(s: str) -> Union[Currency, CurrencyName, str]:
        if s in [x.value for x in Currency]:
            return Currency(s)
        elif s in [x.value for x in CurrencyName]:
            return CurrencyName(s)
        else:
            return s

    return {dt.date.fromisoformat(k): tuple(
            Transaction(decode_inst_tuple(t['portfolio']) if decode_instruments else t['portfolio'],
                        t['portfolio_price'], t['cost'], to_ccy(t['currency']),
                        TransactionDirection(t['direction']) if t['direction'] else None)
            for t in v) for k, v in results.items()}



"""
Copyright 2023 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from typing import Callable, Optional

from gs_quant.session import GsSession


class ApiWithCustomSession:
    __SESSION_SUPPLIER: Optional[Callable[[], GsSession]] = None

    @classmethod
    def set_session_provider(cls, session_supplier: Callable[[], GsSession]):
        """
        To allow session context override specific to this API, set a factory/supplier.
        Default is GsSession.current.
        :param session_supplier: callable which returns a GsSession
        """
        cls.__SESSION_SUPPLIER = session_supplier

    @classmethod
    def set_session(cls, session: GsSession):
        """
        To allow session context override specific to this API, set a session directly.
        Default is GsSession.current.
        :param session: a GsSession
        """
        cls.__SESSION_SUPPLIER = None if session is None else lambda: session

    @classmethod
    def get_session(cls) -> GsSession:
        if cls.__SESSION_SUPPLIER:
            return cls.__SESSION_SUPPLIER()
        else:
            return GsSession.current



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime
import datetime as dt
import logging
import math
from time import time
from typing import Dict, List, Union, Optional

import numpy as np
import pandas as pd
from pydash import get

from gs_quant.api.gs.assets import GsAssetApi
from gs_quant.api.gs.price import GsPriceApi
from gs_quant.errors import MqValueError, MqRequestError
from gs_quant.markets.position_set_utils import _get_asset_temporal_xrefs, \
    _group_temporal_xrefs_into_discrete_time_ranges, _resolve_many_assets
from gs_quant.models.risk_model_utils import _repeat_try_catch_request
from gs_quant.target.common import Position as CommonPosition, PositionPriceInput, PositionSet as CommonPositionSet, \
    PositionTag, Currency, PositionSetWeightingStrategy, MarketDataFrequency
from gs_quant.target.positions_v2_pricing import PositionsPricingParameters, PositionsRequest, PositionSetRequest, \
    PositionsPricingRequest
from gs_quant.target.price import PriceParameters, PositionSetPriceInput

_logger = logging.getLogger(__name__)


class Position:
    def __init__(self,
                 identifier: str,
                 weight: float = None,
                 quantity: float = None,
                 name: str = None,
                 asset_id: str = None,
                 tags: List[PositionTag] = None):
        self.__identifier = identifier
        self.__weight = weight
        self.__quantity = quantity
        self.__name = name
        self.__asset_id = asset_id
        self.__tags = tags
        self.__restricted, self.__hard_to_borrow = None, None

    def __eq__(self, other) -> bool:
        if not isinstance(other, Position):
            return False
        for prop in ['asset_id', 'weight', 'quantity', 'tags']:
            slf = get(self, prop)
            oth = get(other, prop)
            if not (slf is None and oth is None) and not slf == oth:
                return False
        return True

    def __hash__(self):
        return hash(self.asset_id) ^ hash(self.identifier)

    @property
    def identifier(self) -> str:
        return self.__identifier

    @identifier.setter
    def identifier(self, value: str):
        self.__identifier = value

    @property
    def weight(self) -> float:
        return self.__weight

    @weight.setter
    def weight(self, value: float):
        self.__weight = value

    @property
    def quantity(self) -> float:
        return self.__quantity

    @quantity.setter
    def quantity(self, value: float):
        self.__quantity = value

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, value: str):
        self.__name = value

    @property
    def asset_id(self) -> str:
        return self.__asset_id

    @asset_id.setter
    def asset_id(self, value: str):
        self.__asset_id = value

    @property
    def tags(self) -> List[PositionTag]:
        return self.__tags

    @tags.setter
    def tags(self, value: List[PositionTag]):
        self.__tags = value

    @property
    def hard_to_borrow(self) -> bool:
        return self.__hard_to_borrow

    @hard_to_borrow.setter
    def _hard_to_borrow(self, value: bool):
        self.__hard_to_borrow = value

    @property
    def restricted(self) -> bool:
        return self.__restricted

    @restricted.setter
    def _restricted(self, value: bool):
        self.__restricted = value

    def as_dict(self) -> Dict:
        position_dict = dict(identifier=self.identifier, weight=self.weight,
                             quantity=self.quantity, name=self.name, asset_id=self.asset_id, restricted=self.restricted,
                             tags=self.tags)
        return {k: v for k, v in position_dict.items() if v is not None}

    def to_target(self, common: bool = True) -> Union[CommonPosition, PositionPriceInput]:
        """ Returns Position type defined in target file for API payloads """
        if common:
            tags_as_target = self.tags if self.tags else None
            return CommonPosition(self.asset_id, quantity=self.quantity, tags=tags_as_target)
        return PositionPriceInput(self.asset_id, quantity=self.quantity, weight=self.weight)


class PositionSet:
    """

    Position Sets hold a collection of positions associated with a particular date

    """

    def __init__(self,
                 positions: List[Position],
                 date: datetime.date = datetime.date.today(),
                 divisor: float = None,
                 reference_notional: float = None,
                 unresolved_positions: List[Position] = None,
                 unpriced_positions: List[Position] = None):
        if reference_notional is not None:
            for p in positions:
                if p.weight is None:
                    raise MqValueError('Position set with reference notionals must have weights for every position.')
                if p.quantity is not None:
                    raise MqValueError('Position sets with reference notionals cannot have positions with quantities.')
        self.__positions = positions
        self.__date = date
        self.__divisor = divisor
        self.__reference_notional = reference_notional
        self.__unresolved_positions = unresolved_positions if unresolved_positions is not None else []
        self.__unpriced_positions = unpriced_positions if unpriced_positions is not None else []

    def __eq__(self, other) -> bool:
        if len(self.positions) != len(other.positions):
            return False
        if self.date != other.date:
            return False
        if self.reference_notional != other.reference_notional:
            return False
        positions = self.positions
        positions.sort(key=lambda position: position.asset_id)
        other_positions = other.positions
        other_positions.sort(key=lambda position: position.asset_id)
        for i in range(0, len(positions)):
            if positions[i] != other_positions[i]:
                return False
        return True

    @property
    def positions(self) -> List[Position]:
        return self.__positions

    @positions.setter
    def positions(self, value: List[Position]):
        self.__positions = value

    @property
    def date(self) -> datetime.date:
        return self.__date

    @date.setter
    def date(self, value: datetime.date):
        self.__date = value

    @property
    def divisor(self) -> float:
        return self.__divisor

    @property
    def reference_notional(self) -> float:
        return self.__reference_notional

    @reference_notional.setter
    def reference_notional(self, value: float):
        self.__reference_notional = value

    @property
    def unresolved_positions(self) -> List[Position]:
        return self.__unresolved_positions

    @property
    def unpriced_positions(self) -> List[Position]:
        return self.__unpriced_positions

    def get_positions(self) -> pd.DataFrame:
        """
        Retrieve formatted positions

        :return: DataFrame of positions for position set

        **Usage**

        View position set position info

        **Examples**

        Get position set positions:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.get_positions()

        **See also**

        :func:`get_unresolved_positions` :func:`get_unpriced_positions` :func:`resolve` :func:`price`
        """
        positions = [p.as_dict() for p in self.positions]
        return pd.DataFrame(positions)

    def get_unresolved_positions(self) -> pd.DataFrame:
        """
        Retrieve formatted unresolved positions

        :return: DataFrame of unresolved positions for position set

        **Usage**

        View position set unresolved position info

        **Examples**

        Get position set unresolved positions:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.get_unresolved_positions()

        **See also**

        :func:`get_positions` :func:`get_unpriced_positions` :func:`resolve` :func:`price`
        """
        positions = [p.as_dict() for p in self.unresolved_positions]
        return pd.DataFrame(positions)

    def remove_unresolved_positions(self):
        """
        Remove unresolved positions from your position set

        **Usage**

        Remove unresolved positions from your position set

        **Examples**

        Remove unresolved positions from your position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.remove_unresolved_positions()

        **See also**

        :func:`get_positions` :func:`get_unpriced_positions` :func:`resolve` :func:`price`
        :func:`remove_unpriced_positions` :func:`get_unresolved_positions`
        """
        self.positions = [p for p in self.positions if p.asset_id is not None]
        self.__unresolved_positions = None

    def get_unpriced_positions(self) -> pd.DataFrame:
        """
        Retrieve formatted unpriced positions

        :return: DataFrame of unpriced positions for position set

        **Usage**

        View position set unpriced position info

        **Examples**

        Get position set unpriced positions:

        >>> import datetime as dt
        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', quantity=100), Position(identifier='MSFT UW', quantity=100)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.price()
        >>> position_set.get_unpriced_positions()

        **See also**

        :func:`get_positions` :func:`get_unresolved_positions` :func:`resolve` :func:`price`
        """
        positions = [p.as_dict() for p in self.unpriced_positions]
        return pd.DataFrame(positions)

    def remove_unpriced_positions(self):
        """
        Remove unpriced positions from your position set

        **Usage**

        Remove unpriced positions from your position set

        **Examples**

        Remove unpriced positions from your position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.remove_unpriced_positions()

        **See also**

        :func:`get_positions` :func:`get_unpriced_positions` :func:`resolve` :func:`price`
        :func:`get_unresolved_positions` :func:`remove_unresolved_positions`
        """
        self.__unpriced_positions = None

    def get_restricted_positions(self) -> pd.DataFrame:
        """
        Retrieve formatted RTL positions

        :return: DataFrame of RTL positions for position set

        **Usage**

        View position set RTL position info

        **Examples**

        Get position set RTL positions:

        >>> import datetime as dt
        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', quantity=100), Position(identifier='MSFT UW', quantity=100)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.get_restricted_positions()

        **See also**

        :func:`get_positions` :func:`resolve` :func:`price` :func:`remove_restricted_positions`
        :func:`get_hard_to_borrow_positions` :func:`remove_hard_to_borrow_positions`
        """
        positions = [p.as_dict() for p in self.positions if p.restricted]
        return pd.DataFrame(positions)

    def remove_restricted_positions(self):
        """
        Remove RTL positions from your position set

        **Usage**

        Remove RTL positions from your position set

        **Examples**

        Remove RTL positions from your position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.remove_restricted_positions()

        **See also**

        :func:`get_positions` :func:`resolve` :func:`price` :func:`get_restricted_positions`
        :func:`get_hard_to_borrow_positions` :func:`remove_hard_to_borrow_positions`
        """
        self.positions = [p for p in self.positions if p.restricted is not True]

    def get_hard_to_borrow_positions(self) -> pd.DataFrame:
        """
        Retrieve formatted htb positions

        :return: DataFrame of htb positions for position set

        **Usage**

        View position set htb position info

        **Examples**

        Get position set htb positions:

        >>> import datetime as dt
        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', quantity=100), Position(identifier='MSFT UW', quantity=100)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.price()
        >>> position_set.get_hard_to_borrow_positions()

        **See also**

        :func:`get_positions` :func:`resolve` :func:`price` :func:`get_restricted_positions`
        :func:`remove_restricted_positions` :func:`remove_hard_to_borrow_positions`
        """
        positions = [p.as_dict() for p in self.positions if p.hard_to_borrow]
        return pd.DataFrame(positions)

    def remove_hard_to_borrow_positions(self):
        """
        Remove hard to borrow positions from your position set

        **Usage**

        Remove hard to borrow positions from your position set

        **Examples**

        Remove hard to borrow positions from your position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.price()
        >>> position_set.remove_hard_to_borrow_positions()

        **See also**

        :func:`get_positions` :func:`resolve` :func:`price` :func:`get_restricted_positions`
        :func:`remove_restricted_positions` :func:`get_hard_to_borrow_positions`
        """
        self.positions = [p for p in self.positions if p.hard_to_borrow is not True]

    def equalize_position_weights(self):
        """
        Assigns equal weight to each position in position set

        **Usage**

        Assigns equal weight to each position in position set

        **Examples**

        Assign equal weight to each position in position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.equalize_position_weights()

        **See also**

        :func:`get_positions` :func:`redistribute_weights`
        """
        weight = 1 / len(self.positions)
        equally_weighted_positions = []
        for p in self.positions:
            p.weight = weight
            p.quantity = None
            equally_weighted_positions.append(p)
        self.positions = equally_weighted_positions

    def to_frame(self) -> pd.DataFrame:
        """
        Retrieve formatted position set

        :return: DataFrame of position set info

        **Usage**

        View position set info

        **Examples**

        Retrieve formatted position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', quantity=100), Position(identifier='MSFT UW', quantity=100)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.to_frame()

        **See also**

        :func:`from_frame` :func:`from_dicts` :func:`from_list`
        """
        positions = []
        for p in self.positions:
            position = dict(date=self.date.isoformat())
            if self.divisor is not None:
                position.update(dict(divisor=self.divisor))
            position.update(p.as_dict())
            positions.append(position)
        return pd.DataFrame(positions)

    def resolve(self, **kwargs):
        """
        Resolve any unmapped positions

        **Usage**

        Resolve any unmapped positions

        **Examples**

        Resolve any unmapped positions:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW'), Position(identifier='MSFT UW')]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()

        **See also**

        :func:`get_positions` :func:`get_unresolved_positions` :func:`get_unpriced_positions` :func:`price`
        """
        unresolved_positions = [p.identifier for p in self.positions if p.asset_id is None]
        if len(unresolved_positions):
            [id_map, unresolved_positions] = self.__resolve_identifiers(unresolved_positions, self.date, **kwargs)
            self.__unresolved_positions = [p for p in self.positions if p.identifier in unresolved_positions]
            resolved_positions = []
            for p in self.positions:
                if p.identifier in id_map:
                    asset = get(id_map, p.identifier.replace('.', '\\.'))
                    p.asset_id = get(asset, 'id')
                    p.name = get(asset, 'name')
                    p._restricted = get(asset, 'restricted')
                if p.asset_id is not None:
                    resolved_positions.append(p)
            self.positions = resolved_positions

    def redistribute_weights(self):
        """
        Redistribute position weights proportionally for a one-sided position set

        **Usage**

        Redistribute position weights proportionally for a one-sided position set

        **Examples**

        Redistribute position weights proportionally for a one-sided position set:

        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', weight=0.3), Position(identifier='MSFT UW', weight=0.3)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.redistribute_weights()

        **See also**

        :func:`get_positions` :func:`equalize_position_weights` :func:`get_unpriced_positions` :func:`price`
        """
        total_weight = 0
        new_weights, unweighted = [], []
        for p in self.positions:
            if p.weight is None:
                unweighted.append(p.identifier)
            else:
                total_weight += p.weight
        if len(unweighted):
            raise MqValueError(f'Cannot reweight as some positions are missing weights: {unweighted}')

        weight_to_distribute = 1 - total_weight if total_weight < 0 else total_weight - 1
        for p in self.positions:
            p.weight = p.weight - (p.weight / total_weight) * weight_to_distribute
            p.quantity = None
            new_weights.append(p)
        self.positions = new_weights

    def price(self, currency: Optional[Currency] = Currency.USD,
              use_unadjusted_close_price: bool = True,
              weighting_strategy: Optional[PositionSetWeightingStrategy] = None, **kwargs):
        """
        Fetch positions weights from quantities, or vice versa

        :param currency: Reference notional currency (defaults to USD if not passed in)
        :param use_unadjusted_close_price: Use adjusted or unadjusted close prices (defaults to unadjusted)
        :param weighting_strategy: Quantity or Weighted weighting strategy (defaults based on positions info)
        :param use_tags: Determines if tags are used to index the position response for non-netted positions

        **Usage**

        Fetch positions weights from quantities, or vice versa

        **Examples**

        Fetch position weights from quantities:

        >>> from gs_quant.markets.position_set import Position, PositionSet, PositionSetWeightingStrategy
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', quantity=100), Position(identifier='MSFT UW', quantity=100)]
        >>> position_set = PositionSet(positions=my_positions)
        >>> position_set.resolve()
        >>> position_set.price(weighting_strategy=PositionSetWeightingStrategy.Quantity)

        Fetch position quantities from weights:

        >>> import datetime as dt
        >>> from gs_quant.markets.position_set import Position, PositionSet
        >>>
        >>> my_positions = [Position(identifier='AAPL UW', weight=0.5), Position(identifier='MSFT UW', weight=0.5)]
        >>> position_set = PositionSet(positions=my_positions, date= dt.date(2023, 3, 16), reference_notional=10000000)
        >>> position_set.resolve()
        >>> position_set.price(weighting_strategy=PositionSetWeightingStrategy.Weight)

        **See also**

        :func:`get_unpriced_positions` :func:`get_unresolved_positions` :func:`resolve`
        """
        weighting_strategy = self.__get_default_weighting_strategy(self.positions,
                                                                   self.reference_notional,
                                                                   weighting_strategy)
        positions = self.__convert_positions_for_pricing(self.positions, weighting_strategy)
        price_parameters = PriceParameters(currency=currency,
                                           divisor=self.divisor,
                                           frequency=MarketDataFrequency.End_Of_Day,
                                           target_notional=self.reference_notional,
                                           notional_type='Gross',
                                           pricing_date=self.date,
                                           price_regardless_of_assets_missing_prices=True,
                                           weighting_strategy=weighting_strategy,
                                           use_unadjusted_close_price=use_unadjusted_close_price)

        if 'dataset' in kwargs:
            price_parameters.asset_data_set_id = kwargs['dataset']
            price_parameters.frequency = None

        for k, v in kwargs.items():
            price_parameters.__setattr__(k, v)
        results = GsPriceApi.price_positions(PositionSetPriceInput(positions=positions, parameters=price_parameters))
        position_result_map = {f'{p.asset_id}{self.__hash_position_tag_list(p.tags)}': p for p in results.positions}
        priced_positions, unpriced_positions = [], []
        for p in self.positions:
            asset_key = f'{p.asset_id}{self.__hash_position_tag_list(p.tags)}'
            if asset_key in position_result_map:
                p.weight = position_result_map.get(asset_key).weight
                p.quantity = position_result_map.get(asset_key).quantity
                p._hard_to_borrow = position_result_map.get(asset_key).hard_to_borrow
                priced_positions.append(p)
            else:
                unpriced_positions.append(p)
        self.positions = priced_positions
        self.__unpriced_positions = unpriced_positions

    def to_target(self, common: bool = True) -> Union[CommonPositionSet, List[PositionPriceInput]]:
        """ Returns PostionSet type defined in target file for API payloads """
        positions = tuple(p.to_target(common) for p in self.positions)
        return CommonPositionSet(positions, self.date) if common else list(positions)

    @classmethod
    def from_target(cls, position_set: CommonPositionSet):
        """ Create PostionSet instance from PostionSet type defined in target file """
        positions = position_set.positions
        mqids = [position.asset_id for position in positions]
        position_data = cls.__get_positions_data(mqids)
        converted_positions = []
        for p in positions:
            asset = get(position_data, p.asset_id)
            tags = p.tags if p.tags else None
            position = Position(identifier=get(asset, 'bbid'), name=get(asset, 'name'),
                                asset_id=p.asset_id, quantity=p.quantity, tags=tags)
            converted_positions.append(position)
        return cls(converted_positions, position_set.position_date, position_set.divisor)

    @classmethod
    def from_list(cls, positions: List[str], date: datetime.date = datetime.date.today()):
        """
        Create equally-weighted PostionSet instance from a list of identifiers

        **Usage**

        Create equally-weighted PostionSet instance from a list of identifiers

        **Examples**

        Create equally-weighted PostionSet instance from a list of identifiers:

        >>> from gs_quant.markets.position_set import PositionSet
        >>>
        >>> identifiers = ['AAPL UW', 'MSFT UW']
        >>> position_set = PositionSet.from_list(positions=identifiers)

        **See also**

        :func:`get_positions` :func:`resolve` :func:`from_dicts` :func:`from_frame` :func:`to_frame`
        """
        weight = 1 / len(positions)
        converted_positions = [Position(identifier=p, weight=weight) for p in positions]
        return cls(converted_positions, date)

    @classmethod
    def from_dicts(cls, positions: List[Dict],
                   date: datetime.date = datetime.date.today(),
                   reference_notional: float = None,
                   add_tags: bool = False):
        """
        Create PostionSet instance from a list of position-object-like dictionaries

        **Usage**

        Create PostionSet instance from a list of position-object-like dictionaries

        **Examples**

        Create PostionSet instance from a list of position-object-like dictionaries:

        >>> from gs_quant.markets.position_set import PositionSet
        >>>
        >>> my_positions = [{'identifier': 'AAPL UW', 'weight': 0.5}, {'identifier': 'AAPL UW', 'weight': 0.5}]
        >>> position_set = PositionSet.from_dicts(positions=my_positions)

        **See also**

        :func:`get_positions` :func:`resolve` :func:`from_list` :func:`from_frame` :func:`to_frame`
        """
        positions_df = pd.DataFrame(positions)
        return cls.from_frame(positions_df, date, reference_notional, add_tags)

    @classmethod
    def from_frame(cls,
                   positions: pd.DataFrame,
                   date: datetime.date = datetime.date.today(),
                   reference_notional: float = None,
                   add_tags: bool = False):
        """
        Create PostionSet instance from a dataframe of positions

        **Usage**

        Create PostionSet instance from a dataframe of positions

        **Examples**

        Create PostionSet instance from a dataframe of positions:

        >>> import pandas as pd
        >>> from gs_quant.markets.position_set import PositionSet
        >>>
        >>> my_positions = [{'identifier': 'AAPL UW', 'weight': 0.5}, {'identifier': 'AAPL UW', 'weight': 0.5}]
        >>> positions_df = pd.DataFrame(my_positions)
        >>> position_set = PositionSet.from_frame(positions=positions_df)

        **See also**

        :func:`get_positions` :func:`resolve` :func:`from_list` :func:`from_dicts` :func:`to_frame`
        """
        positions.columns = cls.__normalize_position_columns(positions)
        tag_columns = cls.__get_tag_columns(positions) if add_tags else []
        positions = positions[~positions['identifier'].isnull()]
        equalize = not ('quantity' in positions.columns.str.lower() or 'weight' in positions.columns.str.lower())
        equal_weight = 1 / len(positions)

        positions_list = []
        for row in positions.to_dict(orient='records'):
            positions_list.append(
                Position(
                    identifier=row.get('identifier'),
                    asset_id=row.get('id'),
                    name=row.get('name'),
                    weight=equal_weight if equalize else row.get('weight'),
                    quantity=None if equalize else row.get('quantity'),
                    tags=list(PositionTag(tag, get(row, tag)) for tag in tag_columns) if len(tag_columns) else None
                )
            )

        return cls(positions_list, date, reference_notional=reference_notional)

    @staticmethod
    def __get_tag_columns(positions: pd.DataFrame) -> List[str]:
        return [c for c in positions.columns if c.lower() not in ['identifier', 'id', 'quantity', 'weight', 'date']]

    @staticmethod
    def __normalize_position_columns(positions: pd.DataFrame) -> List[str]:
        columns = []
        for c in positions.columns:
            columns.append(c.lower() if c.lower() in ['identifier', 'id', 'quantity', 'weight', 'date'] else c)
        return columns

    @staticmethod
    def __resolve_identifiers(identifiers: List[str], date: datetime.date, **kwargs) -> List:
        unmapped_assets = []
        id_map = {}
        batch_size = 500
        logging.debug(f'Resolving positions in {len(identifiers)/batch_size} batches')
        for i in range(0, len(identifiers), batch_size):
            identifier_batch = identifiers[i: i + batch_size]
            response = GsAssetApi.resolve_assets(
                identifier=identifier_batch,
                fields=['name', 'id', 'tradingRestriction'],
                limit=1,
                as_of=date,
                **kwargs
            )

            for identifier in response:
                if response[identifier] is not None and len(response[identifier]) > 0:
                    id_map[identifier] = {'id': response[identifier][0]['id'],
                                          'name': response[identifier][0]['name'],
                                          'restricted': response[identifier][0].get('tradingRestriction')}
                else:
                    unmapped_assets.append(identifier)

        if len(unmapped_assets) > 0:
            logging.info(f'Error in resolving the following identifiers: {unmapped_assets}. Sifting them out and '
                         f'resolving the rest...')

        return [id_map, unmapped_assets]

    @staticmethod
    def __get_positions_data(mqids: List[str]) -> Dict:
        response = GsAssetApi.get_many_assets_data(id=mqids, fields=['id', 'name', 'bbid'])
        data = {}
        for asset in response:
            data[get(asset, 'id')] = dict(name=get(asset, 'name'), bbid=get(asset, 'bbid'))
        return data

    @staticmethod
    def __get_default_weighting_strategy(positions: List[Position],
                                         reference_notional: float = None,
                                         weighting_strategy: Optional[PositionSetWeightingStrategy] = None
                                         ) -> PositionSetWeightingStrategy:
        missing_weights = [p.identifier for p in positions if p.weight is None]
        missing_quantities = [p.identifier for p in positions if p.quantity is None]
        if weighting_strategy is None:
            if len(missing_weights) and len(missing_quantities):
                raise MqValueError(f'Unable to determine weighting strategy due to missing weights for \
                {missing_weights} and missing quantities for {missing_quantities}')
            if not len(missing_weights) and (reference_notional is not None or len(missing_quantities)):
                weighting_strategy = PositionSetWeightingStrategy.Weight
            else:
                weighting_strategy = PositionSetWeightingStrategy.Quantity
        use_weight = weighting_strategy == PositionSetWeightingStrategy.Weight
        if (use_weight and len(missing_weights)) or (not use_weight and len(missing_quantities)):
            raise MqValueError(f'You must input a {weighting_strategy.value} for the following positions: \
            {missing_weights if use_weight else missing_quantities}')
        if use_weight and reference_notional is None:
            raise MqValueError('You must specify a reference notional in order to price by weight.')
        return weighting_strategy

    @staticmethod
    def __convert_positions_for_pricing(positions: List[Position],
                                        weighting_strategy: PositionSetWeightingStrategy) -> List[PositionPriceInput]:
        position_inputs, missing_ids = [], []
        use_weight = weighting_strategy == PositionSetWeightingStrategy.Weight
        for p in positions:
            if p.asset_id is None:
                missing_ids.append(p.identifier)
            else:
                position_inputs.append(PositionPriceInput(asset_id=p.asset_id,
                                                          weight=p.weight if use_weight else None,
                                                          quantity=None if use_weight else p.quantity,
                                                          tags=p.tags))
        if len(missing_ids):
            raise MqValueError(f'Positions: {missing_ids} are missing asset ids. Resolve your position \
            set or remove unmapped identifiers.')
        return position_inputs

    @staticmethod
    def __hash_position_tag_list(position_tags: List[PositionTag]) -> str:
        hashed_results = ''
        if position_tags is not None:
            for tag in position_tags:
                hashed_results = hashed_results + tag.name + '-' + tag.value
        return hashed_results

    @staticmethod
    def to_frame_many(position_sets: List['PositionSet']) -> pd.DataFrame:
        """Returns dataframe of position sets"""
        position_sets = pd.DataFrame(position_sets, columns=["position_sets"])

        for field in ['date', 'divisor', 'reference_notional']:
            position_sets[field] = [getattr(pos, field, None) for pos in position_sets['position_sets']]

        position_sets['positions'] = [pos.positions for pos in position_sets['position_sets']]

        position_sets = position_sets[position_sets['positions'].apply(lambda x: len(x) > 0)]

        position_sets = position_sets.explode('positions')
        position_sets['positions'] = [pos.as_dict() for pos in position_sets['positions']]

        for field in ['name', 'asset_id', 'identifier', 'weight', 'restricted', 'quantity', 'tags']:
            position_sets[field] = [pos.get(field) for pos in position_sets['positions']]

        columns_to_drop = ["position_sets", "positions"]
        position_sets.drop(columns=columns_to_drop, inplace=True)
        return position_sets

    @staticmethod
    @np.vectorize
    def __build_positions_from_frame(names: pd.Series = None,
                                     identifiers: pd.Series = None,
                                     asset_ids: pd.Series = None,
                                     weights: pd.Series = None,
                                     quantities: pd.Series = None,
                                     restricted: pd.Series = None,
                                     hard_to_borrow: pd.Series = None,
                                     tags: pd.Series = None):
        position = Position(asset_id=asset_ids,
                            name=names,
                            identifier=identifiers,
                            weight=weights if weights else None,
                            quantity=quantities if quantities else None,
                            tags=tags)

        position._restricted = restricted
        position._hard_to_borrow = hard_to_borrow

        return position

    @classmethod
    def resolve_many(cls, position_sets: List['PositionSet'], **kwargs):
        """
        Resolve positions on each holding date into Marquee assets. Positions sets will be updated inplace.
        Each resolved position will have a unique Marquee ID.

            :param position_sets: Positions sets in a list.
            :param kwargs: Additional parameters to send to the GS Resolver API.

            **Usage**

            >>> from gs_quant.markets.position_set import PositionSet, Position, PositionTag
            >>> import datetime as dt
            >>> import pandas as pd

            The input to the function can be a list of `PositionSet` object.

            >>> position_set_list = [
            ...          PositionSet(date=dt.date(2024, 5, 1),
            ...                      reference_notional=1000,
            ...                      positions=[Position(identifier='GS UN',
            ...                                          weight=0.5,
            ...                                          tags=[PositionTag(name="tag1", value="tagvalue1")]),
            ...                                 Position(identifier='AAPL UW',
            ...                                          weight=0.5,
            ...                                          ags=[PositionTag(name="tag2", value="tagvalue2")])]),
            ...          PositionSet(date=dt.date(2024, 5, 1),
            ...                      reference_notional=1000,
            ...                      positions=[Position(identifier='GS UN',
            ...                                          weight=0.5,
            ...                                          tags=[PositionTag(name="tag1", value="tagvalue1")]),
            ...                                 Position(identifier='AAPL UW',
            ...                                           weight=0.5,
            ...                                           tags=[PositionTag(name="tag2", value="tagvalue2")])])
            ...                        ]

            >>> PositionSet.resolve_many(position_set_list)

            **See also**

            :func:`price_many`

            """

        position_sets_df = cls.to_frame_many(position_sets)
        if "name" in position_sets_df.columns.tolist():
            position_sets_df = position_sets_df.drop(columns="name")
        if "asset_id" in position_sets_df.columns.tolist():
            position_sets_df = position_sets_df.drop(columns="asset_id")
        position_sets_df = position_sets_df.dropna(how='all', axis=1)

        position_sets_attributes = position_sets_df.columns.tolist()
        if "quantity" in position_sets_attributes and "weight" in position_sets_attributes:
            raise MqValueError("Cannot have both weight and quantity in position sets")

        asset_temporal_xrefs_df, asset_identifier_type = \
            _get_asset_temporal_xrefs(position_sets_df)
        _group_temporal_xrefs_into_discrete_time_ranges(asset_temporal_xrefs_df)
        resolved_assets_results_df = _resolve_many_assets(asset_temporal_xrefs_df, asset_identifier_type, **kwargs)

        position_sets_df = pd.merge(position_sets_df,
                                    resolved_assets_results_df[["assetId", asset_identifier_type, "name", "asOfDate",
                                                                "tradingRestriction", "startDate", "endDate"]],
                                    how="left",
                                    left_on="identifier",
                                    right_on=asset_identifier_type)
        position_sets_df["date"] = pd.to_datetime(position_sets_df["date"])

        # Fill N/A for startDate and endDate so they are not filtered out (for instance if some positions
        # did not have xrefs or were not resolved. These should be included in the unresolved positions group"
        position_sets_df['startDate'] = position_sets_df['startDate'].fillna(position_sets_df['date'])
        position_sets_df['endDate'] = position_sets_df['endDate'].fillna(position_sets_df['date'])

        position_sets_df = position_sets_df[(position_sets_df["startDate"] <= position_sets_df["date"]) &
                                            (position_sets_df["date"] <= position_sets_df["endDate"])]
        position_sets_df = (
            position_sets_df.drop(columns=[asset_identifier_type, "asOfDate", "startDate", "endDate"])
                            .rename(columns={"tradingRestriction": "restricted"})
                            .fillna(np.nan)
                            .replace([np.nan], [None])
        )

        # Build position sets
        if 'reference_notional' in position_sets_df.columns.tolist():
            if "quantity" in position_sets_df.columns.tolist():
                position_sets_df = position_sets_df.drop(columns='quantity')

        weights_df = position_sets_df['weight'] \
            if 'weight' in position_sets_df.columns.tolist() else None
        quantities_df = position_sets_df['quantity'] \
            if 'quantity' in position_sets_df.columns.tolist() else None
        tags_df = position_sets_df['tags'] if 'tags' in position_sets_df.columns.tolist() else None

        all_positions = cls.__build_positions_from_frame(names=position_sets_df['name'],
                                                         identifiers=position_sets_df['identifier'],
                                                         asset_ids=position_sets_df['assetId'],
                                                         weights=weights_df,
                                                         quantities=quantities_df,
                                                         restricted=position_sets_df['restricted'],
                                                         tags=tags_df)

        position_sets_df['positions'] = all_positions
        position_sets_grouped_by_date = position_sets_df.groupby('date')
        for position_set in position_sets:
            if not isinstance(position_set.date, dt.date):
                position_set.date = pd.Timestamp(position_set.date).to_pydatetime().date()
            positions_on_holding_date_df = position_sets_grouped_by_date.get_group(position_set.date)
            position_set.positions = positions_on_holding_date_df.loc[~positions_on_holding_date_df['assetId'].isnull(),
                                                                      'positions'].tolist()
            unresolved_positions = positions_on_holding_date_df.loc[positions_on_holding_date_df['assetId'].isnull(),
                                                                    'positions'].tolist()
            if unresolved_positions:
                position_set.__unresolved_positions = unresolved_positions

    @classmethod
    def price_many(cls,
                   position_sets: List['PositionSet'],
                   currency: Optional[Currency] = Currency.USD,
                   weighting_strategy: PositionSetWeightingStrategy = None,
                   carryover_positions_for_missing_dates: bool = False,
                   should_reweight: bool = False,
                   allow_fractional_shares: bool = False,
                   allow_partial_pricing: bool = False,
                   batch_size: int = 20,
                   **kwargs):
        """Fetch position weights from quantities or vice versa for a list of position sets. This function modifies the
         input position sets inplace

            :param position_sets: Positions sets in a list.
            :param currency: Currency to use to price. Defaults to USD.
            :param weighting_strategy: The weighting strategy to use. Should be weight or quantity. If None, infers
            weighting strategy from positions metadata
            :param carryover_positions_for_missing_dates: Broadcast previous positions onto dates with missing
            positions. Defaults to False
            :param should_reweight: Ensures total weight across positions on a holding date equals 1. Defaults to False
            :param allow_fractional_shares: Whether to allow fractional shares. Defaults to False
            :param allow_partial_pricing: whether to price a subset of positions in case of errors
            :param batch_size: Size of position sets to send to the pricing API per request. Defaults to 30
            :param kwargs: Additional parameters to pass to the GS pricing API

            **Usage**

            The function expects a required input of positions sets, which can be a dataframe of the format below or
            a list of `PositionSet` object

            >>> from gs_quant.markets.position_set import PositionSet, Position, PositionTag
            >>> import datetime as dt
            >>> import pandas as pd

            The input to the function is a list of `PositoinSet`

            >>> position_set_list = [
            ...     PositionSet(date=dt.date(2024, 5, 1),
            ...                 reference_notional=1000,
            ...                 positions=[Position(identifier='GS UN',
            ...                                      weight=0.5,
            ...                                      tags=[PositionTag(name="tag1", value="tagvalue1")]),
            ...                            Position(identifier='AAPL UW',
            ...                                     weight=0.5,
            ...                                     tags=[PositionTag(name="tag2", value="tagvalue2")])]),
            ...     PositionSet(date=dt.date(2024, 5, 1),
            ...                 reference_notional=1000,
            ...                 positions=[Position(identifier='GS UN',
            ...                                     weight=0.5,
            ...                                     tags=[PositionTag(name="tag1", value="tagvalue1")]),
            ...                            Position(identifier='AAPL UW',
            ...                                     weight=0.5,
            ...                                     tags=[PositionTag(name="tag2", value="tagvalue2")])])
            ...                        ]

            >>> PositionSet.price_many(position_set_list)

            :func:`resolve_many`

            """
        position_sets_to_price_df = cls.to_frame_many(position_sets)
        position_sets_to_price_df = position_sets_to_price_df.dropna(how='all', axis=1)
        position_sets_column_attributes = position_sets_to_price_df.columns.tolist()

        if "quantity" in position_sets_column_attributes and "weight" in position_sets_column_attributes:
            raise MqValueError("Cannot have both weight and quantity in position sets")

        if not weighting_strategy:
            weighting_strategy = PositionSetWeightingStrategy.Weight \
                if "weight" in position_sets_column_attributes \
                   and "reference_notional" in position_sets_column_attributes \
                else PositionSetWeightingStrategy.Quantity

        if weighting_strategy not in [PositionSetWeightingStrategy.Quantity, PositionSetWeightingStrategy.Weight]:
            raise MqValueError("Can only specify a weighting strategy of weight or quantity")

        if weighting_strategy == PositionSetWeightingStrategy.Quantity and \
                "quantity" not in position_sets_column_attributes:
            raise MqValueError("Unable to price positions without position weights and daily reference notional "
                               "or position quantities")

        position_pricing_parameters = PositionsPricingParameters(
            currency=currency.value,
            weighting_strategy=weighting_strategy.value,
            carryover_positions_for_missing_dates=carryover_positions_for_missing_dates,
            should_reweight=should_reweight,
            allow_fractional_shares=allow_fractional_shares
        )

        if kwargs:
            [setattr(position_pricing_parameters, arg, value) for arg, value in kwargs.items()]

        if weighting_strategy == PositionSetWeightingStrategy.Weight:
            positions_with_missing_weights = position_sets_to_price_df[position_sets_to_price_df['weight'].isna()]
            if not positions_with_missing_weights.empty:
                _logger.warning("Some positions do not have weights. These will be filtered out")
        else:
            positions_with_missing_quantities = position_sets_to_price_df[position_sets_to_price_df['quantity'].isna()]
            if not positions_with_missing_quantities.empty:
                _logger.warning("Some positions do not have quantities. These will be filtered out")

        if "weight" not in position_sets_column_attributes:
            position_sets_to_price_df['weight'] = None
        elif "quantity" not in position_sets_column_attributes:
            position_sets_to_price_df["quantity"] = None

        position_sets_to_price_df['positions'] = \
            np.vectorize(
                lambda asset_id, weight, quantity:
                PositionsRequest(asset_id=asset_id, weight=weight, quantity=quantity))(
                position_sets_to_price_df['asset_id'],
                position_sets_to_price_df['weight'],
                position_sets_to_price_df['quantity'])

        # build positionSets requests
        position_sets_grouped_by_date = position_sets_to_price_df.groupby("date")

        all_pos_sets = []
        for date, pos_df in position_sets_grouped_by_date:
            all_pos_sets.append(
                PositionSetRequest(date=date,
                                   positions=pos_df['positions'].tolist(),
                                   target_notional=pos_df['reference_notional'].iat[
                                       0] if weighting_strategy == PositionSetWeightingStrategy.Weight else None))

        batches = np.array_split(all_pos_sets, math.ceil(len(all_pos_sets) / batch_size))
        all_pricing_results = []

        start = time()
        for batch_idx, batch in enumerate(batches):
            _logger.info(f"Pricing batch {batch_idx} of {len(batches)}")
            try:
                payload = PositionsPricingRequest(parameters=position_pricing_parameters,
                                                  position_sets=tuple(batch))
                pricing_results = _repeat_try_catch_request(GsPriceApi.price_many_positions, number_retries=3,
                                                            return_result=True,
                                                            verbose=False,
                                                            pricing_request=payload)
                all_pricing_results += pricing_results
            except MqRequestError as request_exception:
                earliest_pos_set_in_batch = batch[0]
                latest_pos_set_in_batch = batch[-1]
                _logger.error(f"An error occurred while pricing positions on holding dates "
                              f"{earliest_pos_set_in_batch.date} to {latest_pos_set_in_batch.date}: "
                              f"{request_exception}.Consider batching position sets or reducing the batch"
                              f"size")
                if not allow_partial_pricing:
                    raise request_exception
        _logger.info(f"Total time to price positions is {time() - start} seconds")

        next_start = time()
        date_to_priced_position_sets = {
            dt.datetime.strptime(pos_set.get('date'), '%Y-%m-%d').date(): pos_set for pos_set in all_pricing_results}

        for input_position_set in position_sets:
            if not isinstance(input_position_set.date, dt.date):
                # use pandas to infer format of date
                input_position_set.date = pd.to_datetime(input_position_set.date).date()
            if not date_to_priced_position_sets.get(input_position_set.date):
                input_position_set.__unpriced_positions = list(input_position_set.positions)
                input_position_set.positions = None
                continue

            priced_position_set = date_to_priced_position_sets.get(input_position_set.date)
            position_date = dt.datetime.strptime(priced_position_set.get('date'), '%Y-%m-%d').date()
            priced_positions_df = pd.DataFrame(priced_position_set.get('positions'))
            column_from_initial_position_sets_to_merge_by = "weight" \
                if weighting_strategy == PositionSetWeightingStrategy.Weight else "quantity"
            column_from_priced_positions_results_to_merge_by = "referenceWeight" \
                if column_from_initial_position_sets_to_merge_by == "weight" else "quantity"
            priced_positions_df = priced_positions_df\
                .drop_duplicates(subset=['assetId', column_from_priced_positions_results_to_merge_by])

            priced_positions_df = pd.merge(
                position_sets_to_price_df.loc[position_sets_to_price_df['date'] == position_date, :],
                priced_positions_df,
                how="left",
                left_on=['asset_id', column_from_initial_position_sets_to_merge_by],
                right_on=['assetId', column_from_priced_positions_results_to_merge_by],
                suffixes=("original", None)
            )

            unpriced_positions_df = priced_positions_df[priced_positions_df['weight'].isna()] \
                if weighting_strategy == PositionSetWeightingStrategy.Weight else \
                priced_positions_df[priced_positions_df['quantity'].isna()]
            priced_positions_df = priced_positions_df[~priced_positions_df['weight'].isna()] \
                if weighting_strategy == PositionSetWeightingStrategy.Weight else \
                priced_positions_df[~priced_positions_df["quantity"].isna()]
            positions = []
            for record in priced_positions_df.to_dict('records'):
                curr_position = Position(asset_id=record.get('assetId'),
                                         identifier=record.get('identifier'),
                                         name=record.get('name'),
                                         weight=record.get('weight'),
                                         quantity=record.get('quantity'),
                                         tags=record.get('tags'))
                curr_position._restricted = record.get('restricted')
                positions.append(curr_position)

            unpriced_positions = [Position(asset_id=unpriced_record.get('assetId'),
                                           identifier=unpriced_record.get('identifier'),
                                           name=unpriced_record.get('name')) for unpriced_record in
                                  unpriced_positions_df.to_dict('records')]

            input_position_set.positions = positions
            input_position_set._unpriced_positions = unpriced_positions

        _logger.info(f"Total time to process pricing results is {time() - next_start} seconds")



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt

from dataclasses import dataclass, field
from typing import Optional, Union, Tuple

from dataclasses_json import dataclass_json, LetterCase, config

from gs_quant.api.gs.backtests_xasset.json_encoders.request_encoders import legs_encoder, legs_decoder
from gs_quant.api.gs.backtests_xasset.response_datatypes.backtest_datatypes import DateConfig, Trade, \
    CostPerTransaction, Configuration
from gs_quant.api.gs.backtests_xasset.response_datatypes.generic_backtest_datatypes import Strategy
from gs_quant.common import RiskMeasure
from gs_quant.json_convertors import decode_optional_date, decode_date_tuple, encode_date_tuple
from gs_quant.json_convertors_common import encode_risk_measure_tuple, decode_risk_measure_tuple
from gs_quant.priceable import PriceableImpl
from gs_quant.target.backtests import FlowVolBacktestMeasure


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class RiskRequest:
    start_date: Optional[dt.date] = field(default=None, metadata=config(decoder=decode_optional_date))
    end_date: Optional[dt.date] = field(default=None, metadata=config(decoder=decode_optional_date))
    additional_dates: Optional[Tuple[dt.date, ...]] = field(default=None, metadata=config(encoder=encode_date_tuple,
                                                                                          decoder=decode_date_tuple))
    legs: Optional[Tuple[PriceableImpl, ...]] = field(default=None, metadata=config(encoder=legs_encoder,
                                                                                    decoder=legs_decoder))
    measures: Optional[Tuple[RiskMeasure, ...]] = field(default=None,
                                                        metadata=config(encoder=encode_risk_measure_tuple,
                                                                        decoder=decode_risk_measure_tuple))


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class BasicBacktestRequest:
    dates: DateConfig
    trades: Tuple[Trade, ...]
    measures: Tuple[FlowVolBacktestMeasure, ...]
    delta_hedge_frequency: Optional[str] = None
    transaction_costs: Optional[CostPerTransaction] = None
    configuration: Optional[Configuration] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class GenericBacktestRequest:
    strategy: Strategy
    dates: Union[DateConfig, Tuple[dt.date, ...]]
    configuration: Optional[Configuration] = None



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
from typing import Dict, Any, Iterable, Union

from dataclasses_json.cfg import _GlobalConfig

from gs_quant.base import CustomComments
from gs_quant.target.workflow_quote import HedgeTypes
from gs_quant.workflow import VisualStructuringReport, BinaryImageComments, HyperLinkImageComments, \
    CustomDeltaHedge, DeltaHedge

global_config = _GlobalConfig()


def quote_report_from_dict(quote_report_dict: Union[Dict[str, Any], VisualStructuringReport]):
    if quote_report_dict is not None:
        if isinstance(quote_report_dict, VisualStructuringReport):
            return quote_report_dict
        type = quote_report_dict.get('reportType')
        if 'VisualStructuringReport' == type:
            report = VisualStructuringReport.from_dict(quote_report_dict)
            return report
    return None


def quote_reports_from_dicts(quote_report_dicts: Iterable[Dict[str, Any]]):
    if quote_report_dicts is not None:
        reports = []
        for quote_report_dict in quote_report_dicts:
            report = quote_report_from_dict(quote_report_dict)
            reports.append(report)
        return reports
    return None


def custom_comment_from_dict(in_dict: Union[Dict[str, Any], CustomComments]):
    if in_dict is not None:
        if isinstance(in_dict, CustomComments):
            return in_dict
        type = in_dict.get('commentType')
        if 'binaryImageComments' == type:
            out = BinaryImageComments.from_dict(in_dict)
            return out
        if 'hyperLinkImageComments' == type:
            out = HyperLinkImageComments.from_dict(in_dict)
            return out
    return None


def custom_comments_from_dicts(in_dicts: Iterable[Dict[str, Any]]):
    if in_dicts is not None:
        comments = []
        for in_dict in in_dicts:
            report = custom_comment_from_dict(in_dict)
            comments.append(report)
        return comments
    return None


def hedge_type_from_dict(hedge_type_dict: Union[Dict[str, Any], HedgeTypes]):
    if hedge_type_dict is not None:
        if isinstance(hedge_type_dict, HedgeTypes):
            return hedge_type_dict
        type = hedge_type_dict.get('type')
        if 'CustomDeltaHedge' == type:
            hedge_type = CustomDeltaHedge.from_dict(hedge_type_dict)
            return hedge_type
        if 'DeltaHedge' == type:
            hedge_type = DeltaHedge.from_dict(hedge_type_dict)
            return hedge_type
    return None


def hedge_type_from_dicts(in_dicts: Iterable[Dict[str, Any]]):
    if in_dicts is not None:
        hedge_types = []
        for in_dict in in_dicts:
            hedge_type = hedge_type_from_dict(in_dict)
            hedge_types.append(hedge_type)
        return hedge_types
    return None



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt

from dataclasses import dataclass, field
from enum import Enum
from typing import Optional, Tuple, Dict, Union, Any

from dataclasses_json import dataclass_json, LetterCase, config

from gs_quant.api.gs.backtests_xasset.json_encoders.request_encoders import legs_decoder
from gs_quant.api.gs.backtests_xasset.json_encoders.response_datatypes.generic_datatype_encoders import \
    decode_daily_portfolio
from gs_quant.instrument import Instrument
from gs_quant.json_convertors import decode_optional_date, encode_date_tuple, decode_date_tuple
from gs_quant.target.backtests import BacktestTradingQuantityType, EquityMarketModel
from gs_quant.common import Currency, CurrencyName, PricingLocation


class TransactionCostModel(Enum):
    Fixed = 'Fixed'


class TransactionDirection(Enum):
    Entry = 'Entry'
    Exit = 'Exit'


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Transaction:
    portfolio: Tuple[Instrument, ...]
    portfolio_price: Optional[float] = None
    cost: Optional[float] = None
    currency: Optional[Union[Currency, CurrencyName, str]] = None
    direction: Optional[TransactionDirection] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AdditionalResults:
    hedges: Optional[Dict[dt.date, Tuple[Instrument, ...]]] = field(default=None,
                                                                    metadata=config(decoder=decode_daily_portfolio))
    hedge_pnl: Optional[Dict[dt.date, float]] = None
    no_of_calculations: Optional[int] = None

    @classmethod
    def from_dict_custom(cls, data: Any, decode_instruments: bool = True):
        if decode_instruments:
            return cls.from_dict(data)
        return AdditionalResults(hedges=decode_daily_portfolio(data['hedges'], decode_instruments),
                                 hedge_pnl=data['hedge_pnl'],
                                 no_of_calculations=data['no_of_calculations'])


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class DateConfig:
    start_date: dt.date = field(default=None, metadata=config(decoder=decode_optional_date))
    end_date: dt.date = field(default=None, metadata=config(decoder=decode_optional_date))
    frequency: str = '1b'
    holiday_calendar: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class Trade:
    legs: Optional[Tuple[Instrument, ...]] = field(default=None, metadata=config(decoder=legs_decoder))
    buy_frequency: str = None
    buy_dates: Optional[Tuple[dt.date, ...]] = field(default=None, metadata=config(encoder=encode_date_tuple,
                                                                                   decoder=decode_date_tuple))
    holding_period: str = None
    exit_dates: Optional[Tuple[dt.date, ...]] = field(default=None, metadata=config(encoder=encode_date_tuple,
                                                                                    decoder=decode_date_tuple))
    quantity: Optional[float] = None
    quantity_type: BacktestTradingQuantityType = BacktestTradingQuantityType.quantity


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class CostPerTransaction:
    cost_model: TransactionCostModel = TransactionCostModel.Fixed
    cost: float = 0.0


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass(unsafe_hash=True, repr=False)
class Configuration:
    market_data_location: Optional[PricingLocation] = None
    market_model: Optional[EquityMarketModel] = None
    cash_accrual: bool = False



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import backoff

from gs_quant.errors import MqRateLimitedError, MqTimeoutError, MqInternalServerError
from gs_quant.session import GsSession
from gs_quant.target.price import *
from gs_quant.target.positions_v2_pricing import *


class GsPriceApi:
    """GS Price API client implementation"""

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(60),
                          MqRateLimitedError,
                          max_tries=5)
    def price_positions(cls, inputs: PositionSetPriceInput) -> PositionSetPriceResponse:
        url = '/price/positions'
        return GsSession.current._post(url, payload=inputs, cls=PositionSetPriceResponse)

    @classmethod
    def price_many_positions(cls, pricing_request: PositionsPricingRequest) -> dict:
        url = '/positions/price/bulk'
        GsSession.current.api_version = "v2"
        pricing_response = GsSession.current._post(url, payload=pricing_request)
        GsSession.current.api_version = "v1"

        positions = pricing_response.get("positions")

        return positions



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import datetime as dt

import dateutil.parser as dup

from gs_quant.api.gs.portfolios import GsPortfolioApi, Portfolio, PositionSet
from gs_quant.session import *
from gs_quant.target.common import Position


def test_get_many_portfolios(mocker):
    id_1 = 'MP1'
    id_2 = 'MP2'

    mock_response = {'results': (
        Portfolio.from_dict({'id': id_1, 'currency': 'USD', 'name': 'Example Port 1'}),
        Portfolio.from_dict({'id': id_2, 'currency': 'USD', 'name': 'Example Port 2'})
    ), 'totalResults': 2}

    expected_response = (
        Portfolio(id=id_1, currency='USD', name='Example Port 1'),
        Portfolio(id=id_2, currency='USD', name='Example Port 2')
    )

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_portfolios()
    GsSession.current._get.assert_called_with('/portfolios?&limit=100', cls=Portfolio)
    assert response == expected_response


def test_get_portfolio(mocker):
    id_1 = 'MP1'
    mock_response = Portfolio(id=id_1, currency='USD', name='Example Port')

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_portfolio(id_1)
    GsSession.current._get.assert_called_with('/portfolios/{id}'.format(id=id_1), cls=Portfolio)
    assert response == mock_response


def test_create_portfolio(mocker):
    id_1 = 'MP1'

    portfolio = Portfolio(id=id_1, currency='USD', name='Example Port')

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_post', return_value=portfolio)

    # run test
    response = GsPortfolioApi.create_portfolio(portfolio)
    GsSession.current._post.assert_called_with('/portfolios', portfolio, cls=Portfolio)
    assert response == portfolio


def test_update_portfolio(mocker):
    id_1 = 'MP1'

    portfolio = Portfolio(id=id_1, currency='USD', name='Example Port Renamed')

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_put', return_value=portfolio)

    # run test
    response = GsPortfolioApi.update_portfolio(portfolio)
    GsSession.current._put.assert_called_with('/portfolios/{id}'.format(id=id_1), portfolio, cls=Portfolio)
    assert response == portfolio


def test_delete_portfolio(mocker):
    id_1 = 'MP1'

    mock_response = "Successfully deleted portfolio."

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_delete', return_value=mock_response)

    # run test
    response = GsPortfolioApi.delete_portfolio(id_1)
    GsSession.current._delete.assert_called_with('/portfolios/{id}'.format(id=id_1))
    assert response == mock_response


def test_get_portfolio_positions(mocker):
    id_1 = 'MP1'
    start_date = dt.date(2019, 2, 18)
    end_date = dt.date(2019, 2, 19)

    mock_response = {'positionSets': (
        {
            'id': 'mock1',
            'positionDate': '2019-02-18',
            'lastUpdateTime': '2019-02-19T12:10:32.401Z',
            'positions': [
                {'assetId': 'MQA123', 'quantity': 0.3},
                {'assetId': 'MQA456', 'quantity': 0.7}
            ]
        },
        {
            'id': 'mock2',
            'positionDate': '2019-02-19',
            'lastUpdateTime': '2019-02-20T05:04:32.981Z',
            'positions': [
                {'assetId': 'MQA123', 'quantity': 0.4},
                {'assetId': 'MQA456', 'quantity': 0.6}
            ]
        }
    )}

    expected_response = (
        PositionSet(
            id_='mock1',
            position_date=start_date,
            last_update_time=dup.parse('2019-02-19T12:10:32.401Z'),
            positions=(
                Position(assetId='MQA123', quantity=0.3),
                Position(assetId='MQA456', quantity=0.7)
            )),
        PositionSet(
            id_='mock2',
            position_date=end_date,
            last_update_time=dup.parse('2019-02-20T05:04:32.981Z'),
            positions=(
                Position(assetId='MQA123', quantity=0.4),
                Position(assetId='MQA456', quantity=0.6)
            ))
    )

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_positions(id_1, start_date, end_date)

    GsSession.current._get.assert_called_with(
        '/portfolios/{id}/positions?type=close&startDate={sd}&endDate={ed}'.format(id=id_1, sd=start_date, ed=end_date))

    assert response == expected_response


def test_get_portfolio_positions_for_date(mocker):
    id_1 = 'MP1'
    date = dt.date(2019, 2, 18)

    mock_response = {'results': (
        PositionSet.from_dict({
            'id': 'mock1',
            'positionDate': '2019-02-18',
            'lastUpdateTime': '2019-02-19T12:10:32.401Z',
            'positions': [
                {'assetId': 'MQA123', 'quantity': 0.3},
                {'assetId': 'MQA456', 'quantity': 0.7}
            ]
        }),
    )}

    expected_response = (
        PositionSet(
            id_='mock1',
            position_date=date,
            last_update_time=dup.parse('2019-02-19T12:10:32.401Z'),
            positions=(
                Position(assetId='MQA123', quantity=0.3),
                Position(assetId='MQA456', quantity=0.7)
            ))
    )

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_positions_for_date(id_1, date)

    GsSession.current._get.assert_called_with(
        '/portfolios/{id}/positions/{d}?type=close'.format(id=id_1, d=date),
        cls=PositionSet)

    assert response == expected_response


def test_get_latest_portfolio_positions(mocker):
    id_1 = 'MP1'
    date = dt.date(2019, 2, 18)

    mock_response = {
        'results': {
            'id': 'mock1',
            'positionDate': '2019-02-18',
            'lastUpdateTime': '2019-02-19T12:10:32.401Z',
            'positions': [
                {'assetId': 'MQA123', 'quantity': 0.3},
                {'assetId': 'MQA456', 'quantity': 0.7}
            ]
        }
    }

    expected_response = PositionSet(
        id_='mock1',
        position_date=date,
        last_update_time=dup.parse('2019-02-19T12:10:32.401Z'),
        positions=(
            Position(assetId='MQA123', quantity=0.3),
            Position(assetId='MQA456', quantity=0.7)
        ))

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_latest_positions(id_1)

    GsSession.current._get.assert_called_with(
        '/portfolios/{id}/positions/last?type=close'.format(id=id_1))

    assert response == expected_response


def test_get_portfolio_position_dates(mocker):
    id_1 = 'MP1'

    mock_response = {'results': ('2019-02-18', '2019-02-19'), 'totalResults': 2}

    expected_response = (dt.date(2019, 2, 18), dt.date(2019, 2, 19))

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_position_dates(id_1)

    GsSession.current._get.assert_called_with('/portfolios/{id}/positions/dates'.format(id=id_1))

    assert response == expected_response


def test_portfolio_positions_data(mocker):
    mock_response = {'results': [
        {
            'underlyingAssetId': 'MA4B66MW5E27UAFU2CD',
            'divisor': 8305900333.262549,
            'quantity': 0.016836826158,
            'positionType': 'close',
            'bbid': 'EXPE UW',
            'assetId': 'MA4B66MW5E27U8P32SB',
            'positionDate': '2019-11-07',
            'assetClassificationsGicsSector': 'Consumer Discretionary',
            'closePrice': 98.29,
            'ric': 'EXPE.OQ'
        },
    ]}

    expected_response = [
        {
            'underlyingAssetId': 'MA4B66MW5E27UAFU2CD',
            'divisor': 8305900333.262549,
            'quantity': 0.016836826158,
            'positionType': 'close',
            'bbid': 'EXPE UW',
            'assetId': 'MA4B66MW5E27U8P32SB',
            'positionDate': '2019-11-07',
            'assetClassificationsGicsSector': 'Consumer Discretionary',
            'closePrice': 98.29,
            'ric': 'EXPE.OQ'
        },
    ]

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_positions_data('portfolio_id', dt.date(2020, 1, 1), dt.date(2021, 1, 1))

    GsSession.current._get.assert_called_with(
        '/portfolios/portfolio_id/positions/data?startDate=2020-01-01&endDate=2021-01-01')

    assert response == expected_response


def test_get_risk_models_by_coverage(mocker):
    mock_response = {'results': [
        {
            'model': "AXUS4S",
            'businessDate': "2021-03-18",
            'percentInModel': 0.9984356667970278
        },
        {
            'model': "AXUS4M",
            'businessDate': "2021-03-18",
            'percentInModel': 0.9984356667970278
        }
    ]
    }

    expected_response = [
        {
            'model': "AXUS4S",
            'businessDate': "2021-03-18",
            'percentInModel': 0.9984356667970278
        },
        {
            'model': "AXUS4M",
            'businessDate': "2021-03-18",
            'percentInModel': 0.9984356667970278
        }
    ]

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_risk_models_by_coverage('portfolio_id')

    GsSession.current._get.assert_called_with('/portfolios/portfolio_id/models?sortByTerm=Medium')

    assert response == expected_response


def test_get_portfolio_analyze(mocker):
    mock_response = {
        "positionDate": "2024-05-27",
        "bonds": {
            "summary": {
                "trade": {
                    "numberOfBonds": 1,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "DV01": 223.5730399056,
                    "bvalMidPrice": 97.256412,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "ratingSecondHighest": "AAA",
                    "ratingStandardAndPoors": "AA+",
                    "gsLiquidityScore": 1.75825,
                    "illiquidPercentage": 0.0,
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "didNotQuotePercentage": 0.0,
                    "chargeInDollars": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "weight": 100.0,
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "emissionsIntensityEnterpriseValue": 0.0,
                        "emissionsIntensityRevenue": 0.0
                    }
                },
                "buys": {
                    "numberOfBonds": 1,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "DV01": 223.5730399056,
                    "bvalMidPrice": 97.256412,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "ratingSecondHighest": "AAA",
                    "ratingStandardAndPoors": "AA+",
                    "gsLiquidityScore": 1.75825,
                    "illiquidPercentage": 0.0,
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "didNotQuotePercentage": 0.0,
                    "chargeInDollars": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "weight": 100.0,
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "emissionsIntensityEnterpriseValue": 0.0,
                        "emissionsIntensityRevenue": 0.0
                    }
                },
                "sells": {
                    "numberOfBonds": 0,
                    "faceValue": 0.0,
                    "marketValue": 0.0,
                    "weight": 0.0
                }
            },
            "secondHighestRatingGroups": [
                {
                    "name": "AAA",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "standardAndPoorsRatingGroups": [
                {
                    "name": "AA+",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0,
                    "buyDirectionWeight": 1.0,
                    "sellDirectionWeight": "NaN"
                }
            ],
            "maturityGroups": [
                {
                    "name": "1-3y",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "sectorGroups": [
                {
                    "name": "Information Technology",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0,
                    "buyDirectionWeight": 1.0,
                    "sellDirectionWeight": "NaN"
                }
            ],
            "modifiedDurationGroups": [
                {
                    "name": "1-3y",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "gsLiquidityScoreGroups": [
                {
                    "name": "L1",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "seniorityGroups": [
                {
                    "name": "Senior",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "tradesByTicker": [
                {
                    "id": "MA3VNAYYHWQ52SA5",
                    "ticker": "AAPL",
                    "numberOfTrades": 1,
                    "weight": 100.0,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "gsLiquidityScore": 1.75825,
                    "percentOfIssueOutstanding": 0.044444444444444446
                }
            ],
            "constituents": [
                {
                    "assetId": "MA3VNAYYHWQ52SA5",
                    "name": "AAPL 3.350 02/09/2027",
                    "assetClass": "Credit",
                    "type": "Bond",
                    "isin": "US037833CJ77",
                    "bbid": "AM3834018",
                    "cusip": "037833CJ7",
                    "ticker": "AAPL",
                    "currency": "USD",
                    "region": "Americas",
                    "sector": "Information Technology",
                    "seniority": "Senior",
                    "ratingSecondHighest": "AAA",
                    "ratingSecondHighestLinear": 1,
                    "ratingStandardAndPoors": "AA+",
                    "ratingStandardAndPoorsLinear": 2,
                    "gsLiquidityScore": 1.75825,
                    "gsLiquidityScoreGroup": "L1",
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "maturity": 2.0,
                    "coupon": 3.35,
                    "bvalMidPrice": 97.256412,
                    "dirtyPrice": 97.270928666667,
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "benchmark": "GV91282CKZ3",
                    "quoteConvention": "Spread",
                    "chargeInQuoteConvention": "2.5 bps",
                    "chargeInQuoteConventionTwo": "2.5 bps",
                    "chargeInLocalCurrency": 0.055893259976399996,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "fxSpot": 1.0,
                    "direction": "Buy",
                    "faceValue": 1000000.0,
                    "currentFaceValue": 1000000.0,
                    "accruedInterestStandard": 0.014516666667,
                    "marketValue": 972564.12,
                    "dirtyMarketValue": 972709.28666667,
                    "DV01": 223.5730399056,
                    "weightOfFaceValue": 100.0,
                    "weightOfMarketValue": 100.0,
                    "weightOfDV01": 100.0,
                    "indicativePrice": 97.289,
                    "indicativeQuantity": 5000000.0,
                    "indicativeSpread": 25.0,
                    "indicativeYield": 4.48,
                    "girFactors": {
                        "carry": 1.0,
                        "value": 3.0,
                        "risk": 1.0,
                        "momentum": 1.0,
                        "liquidity": 4.0,
                        "highLeverage": 0.0,
                        "lowLeverage": 0.0,
                        "highInterestCoverage": 0.0,
                        "lowInterestCoverage": 0.0,
                        "cyclical": 1.0,
                        "defensive": 0.0,
                        "commoditiesRelated": 0.0,
                        "tax": 4.0
                    },
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "scienceBasedTarget": "Yes",
                        "netZeroEmissionsTarget": "Yes"
                    },
                    "pbData": {
                        "indicativeShortFinancingLabel": "GC",
                        "indicativeLongFinancingLabel": "IG Funding Rate"
                    },
                    "bbgData": {
                        "countryOfRisk": "US",
                        "paymentRank": "Sr Unsecured",
                        "industrySector": "Technology",
                        "industryGroup": "Computers",
                        "industrySubGroup": "Computers"
                    }
                }
            ]
        }
    }

    expected_response = {
        "positionDate": "2024-05-27",
        "bonds": {
            "summary": {
                "trade": {
                    "numberOfBonds": 1,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "DV01": 223.5730399056,
                    "bvalMidPrice": 97.256412,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "ratingSecondHighest": "AAA",
                    "ratingStandardAndPoors": "AA+",
                    "gsLiquidityScore": 1.75825,
                    "illiquidPercentage": 0.0,
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "didNotQuotePercentage": 0.0,
                    "chargeInDollars": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "weight": 100.0,
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "emissionsIntensityEnterpriseValue": 0.0,
                        "emissionsIntensityRevenue": 0.0
                    }
                },
                "buys": {
                    "numberOfBonds": 1,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "DV01": 223.5730399056,
                    "bvalMidPrice": 97.256412,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "ratingSecondHighest": "AAA",
                    "ratingStandardAndPoors": "AA+",
                    "gsLiquidityScore": 1.75825,
                    "illiquidPercentage": 0.0,
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "didNotQuotePercentage": 0.0,
                    "chargeInDollars": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "weight": 100.0,
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "emissionsIntensityEnterpriseValue": 0.0,
                        "emissionsIntensityRevenue": 0.0
                    }
                },
                "sells": {
                    "numberOfBonds": 0,
                    "faceValue": 0.0,
                    "marketValue": 0.0,
                    "weight": 0.0
                }
            },
            "secondHighestRatingGroups": [
                {
                    "name": "AAA",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "standardAndPoorsRatingGroups": [
                {
                    "name": "AA+",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0,
                    "buyDirectionWeight": 1.0,
                    "sellDirectionWeight": "NaN"
                }
            ],
            "maturityGroups": [
                {
                    "name": "1-3y",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "sectorGroups": [
                {
                    "name": "Information Technology",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0,
                    "buyDirectionWeight": 1.0,
                    "sellDirectionWeight": "NaN"
                }
            ],
            "modifiedDurationGroups": [
                {
                    "name": "1-3y",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "gsLiquidityScoreGroups": [
                {
                    "name": "L1",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "seniorityGroups": [
                {
                    "name": "Senior",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "marketValue": 972564.12,
                    "weight": 100.0
                }
            ],
            "tradesByTicker": [
                {
                    "id": "MA3VNAYYHWQ52SA5",
                    "ticker": "AAPL",
                    "numberOfTrades": 1,
                    "weight": 100.0,
                    "faceValue": 1000000.0,
                    "marketValue": 972564.12,
                    "gsLiquidityScore": 1.75825,
                    "percentOfIssueOutstanding": 0.044444444444444446
                }
            ],
            "constituents": [
                {
                    "assetId": "MA3VNAYYHWQ52SA5",
                    "name": "AAPL 3.350 02/09/2027",
                    "assetClass": "Credit",
                    "type": "Bond",
                    "isin": "US037833CJ77",
                    "bbid": "AM3834018",
                    "cusip": "037833CJ7",
                    "ticker": "AAPL",
                    "currency": "USD",
                    "region": "Americas",
                    "sector": "Information Technology",
                    "seniority": "Senior",
                    "ratingSecondHighest": "AAA",
                    "ratingSecondHighestLinear": 1,
                    "ratingStandardAndPoors": "AA+",
                    "ratingStandardAndPoorsLinear": 2,
                    "gsLiquidityScore": 1.75825,
                    "gsLiquidityScoreGroup": "L1",
                    "percentOfIssueOutstanding": 0.044444444444444446,
                    "maturity": 2.0,
                    "coupon": 3.35,
                    "bvalMidPrice": 97.256412,
                    "dirtyPrice": 97.270928666667,
                    "gSpread": 19.588,
                    "zSpread": 36.634,
                    "bvalUpdateTime": "2024-07-15T20:00:00Z",
                    "yieldToConvention": 4.494023,
                    "modifiedDuration": 2.2988,
                    "spreadToBenchmark": 28.5952,
                    "benchmark": "GV91282CKZ3",
                    "quoteConvention": "Spread",
                    "chargeInQuoteConvention": "2.5 bps",
                    "chargeInQuoteConventionTwo": "2.5 bps",
                    "chargeInLocalCurrency": 0.055893259976399996,
                    "chargeInEntityCurrency": 0.055893259976399996,
                    "chargeInBps": 2.5,
                    "fxSpot": 1.0,
                    "direction": "Buy",
                    "faceValue": 1000000.0,
                    "currentFaceValue": 1000000.0,
                    "accruedInterestStandard": 0.014516666667,
                    "marketValue": 972564.12,
                    "dirtyMarketValue": 972709.28666667,
                    "DV01": 223.5730399056,
                    "weightOfFaceValue": 100.0,
                    "weightOfMarketValue": 100.0,
                    "weightOfDV01": 100.0,
                    "indicativePrice": 97.289,
                    "indicativeQuantity": 5000000.0,
                    "indicativeSpread": 25.0,
                    "indicativeYield": 4.48,
                    "girFactors": {
                        "carry": 1.0,
                        "value": 3.0,
                        "risk": 1.0,
                        "momentum": 1.0,
                        "liquidity": 4.0,
                        "highLeverage": 0.0,
                        "lowLeverage": 0.0,
                        "highInterestCoverage": 0.0,
                        "lowInterestCoverage": 0.0,
                        "cyclical": 1.0,
                        "defensive": 0.0,
                        "commoditiesRelated": 0.0,
                        "tax": 4.0
                    },
                    "girESG": {
                        "gPercentile": 87.88,
                        "gRegionalPercentile": 74.57,
                        "esPercentile": 92.78,
                        "esDisclosurePercentage": 80.95,
                        "esMomentumPercentile": 0.0
                    },
                    "carbon": {
                        "scienceBasedTarget": "Yes",
                        "netZeroEmissionsTarget": "Yes"
                    },
                    "pbData": {
                        "indicativeShortFinancingLabel": "GC",
                        "indicativeLongFinancingLabel": "IG Funding Rate"
                    },
                    "bbgData": {
                        "countryOfRisk": "US",
                        "paymentRank": "Sr Unsecured",
                        "industrySector": "Technology",
                        "industryGroup": "Computers",
                        "industrySubGroup": "Computers"
                    }
                }
            ]
        }
    }

    # mock GsSession
    mocker.patch.object(
        GsSession.__class__,
        'default_value',
        return_value=GsSession.get(
            Environment.QA,
            'client_id',
            'secret'))
    mocker.patch.object(GsSession.current, '_get', return_value=mock_response)

    # run test
    response = GsPortfolioApi.get_portfolio_analyze('portfolio_id')

    GsSession.current._get.assert_called_with('/portfolios/portfolio_id/analyze')

    assert response == expected_response



"""
Copyright 2018 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
import traceback
from contextlib import ContextDecorator
from typing import Tuple, Optional

import pandas as pd
from opentracing import Span, UnsupportedFormatException, SpanContextCorruptedException
from opentracing import Tracer as OpenTracer
from opentracing.mocktracer import MockTracer
from opentracing.scope_managers.contextvars import ContextVarsScopeManager

from gs_quant.errors import *

_logger = logging.getLogger(__name__)


class TracerFactory:
    __tracer_instance = None

    def get(self) -> OpenTracer:
        if TracerFactory.__tracer_instance is None:
            TracerFactory.__tracer_instance = MockTracer(scope_manager=ContextVarsScopeManager())
        return TracerFactory.__tracer_instance


class Tracer(ContextDecorator):
    __factory = TracerFactory()

    @staticmethod
    def get_instance() -> OpenTracer:
        return Tracer.__factory.get()

    @staticmethod
    def inject(format, carrier):
        instance = Tracer.get_instance()
        span = instance.active_span
        if span is not None:
            try:
                instance.inject(span.context, format, carrier)
            except UnsupportedFormatException:
                pass

    @staticmethod
    def extract(format, carrier):
        instance = Tracer.get_instance()
        try:
            return instance.extract(format, carrier)
        except (UnsupportedFormatException, SpanContextCorruptedException):
            pass

    @staticmethod
    def set_factory(factory: TracerFactory):
        Tracer.__factory = factory

    def __init__(self, label: str = 'Execution', print_on_exit: bool = False, threshold: int = None,
                 wrap_exceptions=False):
        self.__print_on_exit = print_on_exit
        self.__label = label
        self.__threshold = threshold
        self.wrap_exceptions = wrap_exceptions

    def __enter__(self):
        self.__scope = Tracer.get_instance().start_active_span(operation_name=self.__label)
        return self.__scope

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_value:
            Span._on_error(self.__scope.span, exc_type, exc_value, Tracer.__format_traceback(exc_type, exc_value))
        self.__scope.close()
        if self.wrap_exceptions and exc_type is not None and not exc_type == MqWrappedError:
            raise MqWrappedError(f'Unable to calculate: {self.__label}') from exc_value

    @staticmethod
    def record_exception(e):
        span = Tracer.get_instance().active_span
        if span is not None:
            try:
                Span._on_error(span, type(e), e, Tracer.__format_traceback(e, type(e)))
            except Exception:
                pass

    @staticmethod
    def __format_traceback(exc_type, exc_value):
        if exc_value is None:
            return ''
        try:
            return ''.join(traceback.format_exception(exc_type, exc_value, None, limit=10))
        except Exception:
            return ''

    @staticmethod
    def reset():
        Tracer.get_instance().reset()

    @staticmethod
    def get_spans():
        return Tracer.get_instance().finished_spans()

    @staticmethod
    def plot(reset=False):
        try:
            import plotly.express as px
        except ImportError:
            _logger.warning('Package "plotly" required to visualise the trace, printing instead')
            Tracer.print(reset)
            return
        color_list = ('#2b76f7', '#5a8efb', '#79aaff', '#89bbff', '#cdddff')
        error_color = 'rgb(244, 127, 114)'
        ordered_spans, _ = Tracer.gather_data(False)
        span_df = pd.DataFrame.from_records([(
            f'#{i}',
            f'{s.operation_name} {int(1000 * (s.finish_time - s.start_time)):,.0f}ms',
            dt.datetime.fromtimestamp(s.start_time),
            dt.datetime.fromtimestamp(s.finish_time),
            '\n '.join([f'{k}={v}' for k, v in s.tags.items()]) if s.tags else '',
        ) for i, (depth, s) in enumerate(ordered_spans)], columns=['id', 'operation', 'start', 'end', 'tags'])
        color_map = {f'#{i}': error_color if 'error' in s.tags else color_list[depth % len(color_list)] for
                     i, (depth, s) in enumerate(ordered_spans)}
        fig = px.timeline(
            data_frame=span_df,
            width=1000,
            height=40 + 30 * len(span_df),
            x_start="start",
            x_end="end",
            y="id",
            hover_data='tags',
            text='operation',
            color='id',
            color_discrete_map=color_map
        )
        fig.update_layout(
            showlegend=False,
            yaxis_visible=False,
            yaxis_showticklabels=False,
            paper_bgcolor='rgba(0,0,0,0)',
            plot_bgcolor='rgba(0,0,0,0)',
            margin={'l': 0, 't': 0, 'b': 0, 'r': 0}
        )
        if reset:
            Tracer.reset()
        fig.show()

    @staticmethod
    def gather_data(as_string: bool = True, root_id: Optional[str] = None):
        spans = Tracer.get_spans()
        spans_by_parent = {}

        for span in reversed(spans):
            spans_by_parent.setdefault(span.parent_id, []).append(span)

        def _build_tree(parent_span, depth):
            if as_string:
                name = f'{"* " * depth}{parent_span.operation_name}'
                elapsed = (parent_span.finish_time - parent_span.start_time) * 1000 if parent_span.finished else 'N/A'
                error = " [Error]" if parent_span.tags.get('error', False) else ""
                lines.append(f'{name:<50}{elapsed:>8.1f} ms{error}')
            else:
                lines.append((depth, parent_span))
            for child_span in reversed(spans_by_parent.get(parent_span.context.span_id, [])):
                _build_tree(child_span, depth + 1)

        total = 0
        lines = []
        # By default, we look for the span with no parent, but this might not always be what we want
        for span in reversed(spans_by_parent.get(root_id, [])):
            _build_tree(span, 0)
            total += (span.finish_time - span.start_time) * 1000

        if as_string:
            tracing_str = '\n'.join(lines)
            return tracing_str, total
        else:
            return lines, total

    @staticmethod
    def print(reset=True, root_id=None):
        tracing_str, total = Tracer.gather_data(root_id=root_id)
        _logger.warning(f'Tracing Info:\n{tracing_str}\n{"-" * 61}\nTOTAL:{total:>52.1f} ms')
        if reset:
            Tracer.reset()
        return tracing_str, total


def parse_tracing_line_args(line: str) -> Tuple[Optional[str], bool]:
    stripped = tuple(s for s in line.split(' ') if s != '')
    if len(stripped) > 0 and stripped[0] in ('chart', 'plot', 'graph'):
        return tuple(stripped[1:]) if len(stripped[1:]) else None, True
    return stripped if len(stripped) else None, False


try:
    # Attempt to import/register some jupyter magic
    import gs_quant_internal.tracing.jupyter  # noqa
except ImportError:
    try:
        from IPython.core.magic import register_cell_magic
        from IPython import get_ipython

        @register_cell_magic("trace")
        def trace_ipython_cell(line, cell):
            """Wraps the execution of a cell in a tracer call and prints"""
            span_name, show_chart = parse_tracing_line_args(line)
            if cell is None:
                return line
            with Tracer(label=span_name):
                res = get_ipython().run_cell(cell)
                if res.error_in_exec:
                    Tracer.record_exception(res.error_in_exec)
            if show_chart:
                Tracer.plot(True)
            else:
                Tracer.print(True)
            return None
    except Exception:
        pass



"""
Copyright 2024 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
'AS IS' BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt

import pandas as pd
import pytest

from gs_quant.errors import MqValueError
from gs_quant.markets.position_set import PositionSet, Position, GsPriceApi, PositionTag, PositionSetWeightingStrategy
import gs_quant.markets.position_set as position_set_module
from copy import deepcopy


def test_position_resolve_many(mocker):
    unresolved_position_sets = [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])]),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])])

    ]

    resolved_positions = [
        {"assetId": "MA4B66MW5E27UAHKG34", "name": "GS", "bbid": "GS UN", "tradingRestriction": True,
         "asOfDate": dt.datetime.strptime("2952-12-31", '%Y-%m-%d'),
         "startDate": dt.datetime.strptime("1952-01-01", '%Y-%m-%d'),
         "endDate": dt.datetime.strptime("2952-12-31", '%Y-%m-%d')},
        {"assetId": "MA4B66MW5E27U9VBB94", "name": "Apple", "bbid": "AAPL UW", "tradingRestriction": False,
         "asOfDate": dt.datetime.strptime("2952-12-31", '%Y-%m-%d'),
         "startDate": dt.datetime.strptime("1952-01-01", '%Y-%m-%d'),
         "endDate": dt.datetime.strptime("2952-12-31", '%Y-%m-%d')}
    ]

    xref_results = [
        {"assetId": "MA4B66MW5E27UAHKG34", "bbid": "GS UN", "delisted": 'no',
         "startDate": "1952-01-01", "endDate": "2952-12-31"},
        {"assetId": "MA4B66MW5E27U9VBB94", "bbid": "AAPL UW", "delisted": 'no',
         "startDate": "1952-01-01", "endDate": "2952-12-31"},
    ]

    mocker.patch.object(position_set_module, "_get_asset_temporal_xrefs",
                        return_value=(pd.DataFrame(xref_results), "bbid"))
    mocker.patch.object(position_set_module, "_resolve_many_assets", return_value=pd.DataFrame(resolved_positions))

    unresolved_position_sets_copy = deepcopy(unresolved_position_sets)
    PositionSet.resolve_many(unresolved_position_sets_copy)

    resolved_position_sets = [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])]),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])])
    ]

    for idx, pos_set in enumerate(unresolved_position_sets_copy):
        assert pos_set == resolved_position_sets[idx]

    # There are unresolved positions
    mocker.patch.object(position_set_module, "_get_asset_temporal_xrefs",
                        return_value=(pd.DataFrame([xref_results[1]]), "bbid"))
    mocker.patch.object(position_set_module, "_resolve_many_assets", return_value=pd.DataFrame([resolved_positions[1]]))

    unresolved_position_sets_copy = deepcopy(unresolved_position_sets)

    PositionSet.resolve_many(unresolved_position_sets_copy)

    expected_position_set = [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])],
                    unresolved_positions=[Position(identifier='GS UN',
                                                   name="GS",
                                                   weight=0.5,
                                                   tags=[PositionTag(name="tag1", value="tagvalue1")])]
                    ),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[
                        Position(identifier='AAPL UW',
                                 asset_id="MA4B66MW5E27U9VBB94",
                                 name="Apple",
                                 weight=0.5,
                                 tags=[PositionTag(name="tag2", value="tagvalue2")])],
                    unresolved_positions=[Position(identifier='GS UN',
                                                   name="GS",
                                                   weight=0.5,
                                                   tags=[PositionTag(name="tag1", value="tagvalue1")])]
                    )
    ]

    unresolved_position_sets_copy.sort(key=lambda x: x.date)
    expected_position_set.sort(key=lambda x: x.date)

    for idx, position_set in enumerate(unresolved_position_sets_copy):
        assert position_set == expected_position_set[idx]
        assert len(position_set.unresolved_positions) == 1
        assert len(expected_position_set[idx].unresolved_positions) == 1

        actual_unresolved_position = position_set.unresolved_positions[0]
        expected_unresolved_position = expected_position_set[idx].unresolved_positions[0]
        assert actual_unresolved_position.__eq__(expected_unresolved_position)


@pytest.fixture()
def position_sets_with_tags_and_notional():
    return [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])]),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")])])
    ]


@pytest.fixture()
def expected_position_pricing_result():
    return [{'date': '2024-04-30',
             'positions': [{'assetId': 'MA4B66MW5E27UAHKG34', 'weight': 0.42671,
                            'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
                            'notional': 1000, 'referenceWeight': 0.5},
                           {'assetId': 'MA4B66MW5E27U9VBB94', 'weight': 0.51099,
                            'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
                            'notional': 1000, 'referenceWeight': 0.5}], 'targetNotional': 1000},
            {'date': '2024-05-01',
             'positions': [{'assetId': 'MA4B66MW5E27UAHKG34', 'weight': 0.42671,
                            'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
                            'notional': 1000, 'referenceWeight': 0.5},
                           {'assetId': 'MA4B66MW5E27U9VBB94', 'weight': 0.51099,
                            'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
                            'notional': 1000, 'referenceWeight': 0.5}],
             'targetNotional': 1000}]


def test_position_price_many(mocker,
                             position_sets_with_tags_and_notional,
                             expected_position_pricing_result):
    # All positions on each holding date are priced

    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=expected_position_pricing_result)
    new_position_set_list = deepcopy(position_sets_with_tags_and_notional)
    PositionSet.price_many(new_position_set_list)

    expected_positions = {dt.date(2024, 4, 30): [Position(identifier='GS UN',
                                                          asset_id="MA4B66MW5E27UAHKG34",
                                                          name="GS",
                                                          weight=0.42671,
                                                          quantity=1,
                                                          tags=[PositionTag(name="tag1", value="tagvalue1")]),
                                                 Position(identifier='AAPL UW',
                                                          asset_id="MA4B66MW5E27U9VBB94",
                                                          name="Apple",
                                                          weight=0.51099,
                                                          quantity=1,
                                                          tags=[PositionTag(name="tag2", value="tagvalue2")])],
                          dt.date(2024, 5, 1): [Position(identifier='GS UN',
                                                         asset_id="MA4B66MW5E27UAHKG34",
                                                         name="GS",
                                                         weight=0.42671,
                                                         quantity=1,
                                                         tags=[PositionTag(name="tag1", value="tagvalue1")]),
                                                Position(identifier='AAPL UW',
                                                         asset_id="MA4B66MW5E27U9VBB94",
                                                         name="Apple",
                                                         weight=0.51099,
                                                         quantity=1,
                                                         tags=[PositionTag(name="tag2", value="tagvalue2")])]}

    [pos_set.positions.sort(key=lambda position: position.asset_id) for pos_set in new_position_set_list]
    [positions.sort(key=lambda position: position.asset_id) for positions in expected_positions.values()]

    for pos_set in new_position_set_list:
        assert pos_set.positions == expected_positions.get(pos_set.date)

    # weighting strategy passed is 'quantity' but positions do not have quantities
    with pytest.raises(MqValueError,
                       match="Unable to price positions without position weights and "
                             "daily reference notional or position quantities"):
        PositionSet.price_many(position_sets_with_tags_and_notional,
                               weighting_strategy=PositionSetWeightingStrategy.Quantity)

    # Invalid weighting strategies
    with pytest.raises(MqValueError,
                       match="Can only specify a weighting strategy of weight or quantity"):
        PositionSet.price_many(position_sets_with_tags_and_notional,
                               weighting_strategy=PositionSetWeightingStrategy.Market_Capitalization)

    # There are unpriced results
    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=[expected_position_pricing_result[1]])
    new_position_set_list = deepcopy(position_sets_with_tags_and_notional)
    PositionSet.price_many(new_position_set_list)

    first_position_set = new_position_set_list[0]
    positions = first_position_set.positions
    unpriced_positions = first_position_set.unpriced_positions

    assert not positions
    for position in unpriced_positions:
        if position.identifier == "GS UN":
            assert position.asset_id == "MA4B66MW5E27UAHKG34"
            assert position.weight == 0.5
            assert not position.quantity
            assert len(position.tags) == 1
            tag = position.tags[0]
            assert tag.name == "tag1"
            assert tag.value == "tagvalue1"

        else:
            assert position.asset_id == "MA4B66MW5E27U9VBB94"
            assert position.weight == 0.5
            assert not position.quantity
            assert len(position.tags) == 1
            tag = position.tags[0]
            assert tag.name == "tag2"
            assert tag.value == "tagvalue2"

    second_position_set = new_position_set_list[1]
    positions = second_position_set.positions
    unpriced_positions = second_position_set.unpriced_positions

    assert not unpriced_positions
    for position in positions:
        if position.identifier == "GS UN":
            assert position.asset_id == "MA4B66MW5E27UAHKG34"
            assert position.weight == 0.42671
            assert position.quantity == 1
            assert len(position.tags) == 1
            tag = position.tags[0]
            assert tag.name == "tag1"
            assert tag.value == "tagvalue1"

        else:
            assert position.asset_id == "MA4B66MW5E27U9VBB94"
            assert position.weight == 0.51099
            assert position.quantity == 1
            assert len(position.tags) == 1
            tag = position.tags[0]
            assert tag.name == "tag2"
            assert tag.value == "tagvalue2"

    # Positions do not have tags
    new_position_set_list = deepcopy(position_sets_with_tags_and_notional)
    position_sets_no_tags = [deepcopy(pos_set) for pos_set in new_position_set_list]
    for position_set_no_tag in position_sets_no_tags:
        [setattr(position, "tags", None) for position in position_set_no_tag.positions]

    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=expected_position_pricing_result)
    PositionSet.price_many(position_sets_no_tags)

    for position_set in position_sets_no_tags:
        priced_positions = position_set.positions
        for position in priced_positions:
            if position.identifier == "GS UN":
                assert position.asset_id == "MA4B66MW5E27UAHKG34"
                assert position.weight == 0.42671
                assert position.quantity == 1
                assert not position.tags

            else:
                assert position.asset_id == "MA4B66MW5E27U9VBB94"
                assert position.weight == 0.51099
                assert position.quantity == 1
                assert not position.tags

    # Positions with different tags and common underlying asset ID
    different_tags_common_asset_id_positions = [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=-0.8,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=None)
                               ]
                    ),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=-0.8,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag3", value="tagvalue3")])
                               ]
                    )
    ]

    same_assets_different_weights_pricing_results = [
        {'date': '2024-04-30',
         'positions': [
             {'assetId': 'MA4B66MW5E27UAHKG34', 'weight': 0.38,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
              'notional': 1000, 'referenceWeight': 0.5},
             {'assetId': 'MA4B66MW5E27UAHKG34', 'weight': -0.68,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': -1.5,
              'notional': -1000, 'referenceWeight': -0.8},
             {'assetId': 'MA4B66MW5E27U9VBB94', 'weight': 0.41,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
              'notional': 1000, 'referenceWeight': 0.5}],
         'targetNotional': 1000},
        {'date': '2024-05-01',
         'positions': [
             {'assetId': 'MA4B66MW5E27UAHKG34', 'weight': 0.42671,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
              'notional': 1000, 'referenceWeight': 0.5},
             {'assetId': 'MA4B66MW5E27UAHKG34', 'weight': -0.72,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': -1.5,
              'notional': -1000, 'referenceWeight': -0.8},
             {'assetId': 'MA4B66MW5E27U9VBB94', 'weight': 0.51099,
              'closePrice': 1000, 'fxClosePrice': 1, 'quantity': 1,
              'notional': 1000, 'referenceWeight': 0.5}],
         'targetNotional': 1000}
    ]
    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=same_assets_different_weights_pricing_results)
    PositionSet.price_many(different_tags_common_asset_id_positions)

    for position_set in different_tags_common_asset_id_positions:
        if position_set.date == dt.date(2024, 4, 30):
            for position in position_set.positions:
                if position.identifier == "GS UN":
                    assert position.asset_id == "MA4B66MW5E27UAHKG34"
                    assert position.name == "GS"
                    assert position.weight in [0.38, -0.68]
                    assert position.quantity in [1, -1.5]

                    if position.weight == 0.38:
                        assert position.tags[0] == PositionTag(name="tag1", value="tagvalue1")
                    else:
                        assert position.tags[0] == PositionTag(name="tag2", value="tagvalue2")
                else:
                    assert position.asset_id == "MA4B66MW5E27U9VBB94"
                    assert position.name == "Apple"
                    assert position.weight == 0.41
                    assert not position.tags

        else:
            for position in position_set.positions:
                if position.identifier == "GS UN":
                    assert position.asset_id == "MA4B66MW5E27UAHKG34"
                    assert position.name == "GS"
                    assert position.weight in [0.42671, -0.72]
                    assert position.quantity in [1, -1.5]

                    if position.weight == 0.42671:
                        assert position.tags[0] == PositionTag(name="tag1", value="tagvalue1")
                    else:
                        assert position.tags[0] == PositionTag(name="tag2", value="tagvalue2")
                else:
                    assert position.asset_id == "MA4B66MW5E27U9VBB94"
                    assert position.name == "Apple"
                    assert position.weight == 0.51099
                    assert position.tags[0] == PositionTag(name="tag3", value="tagvalue3")

    # There are empty positions in some position sets
    different_tags_common_asset_id_positions = [
        PositionSet(date=dt.date(2024, 4, 30),
                    reference_notional=1000,
                    positions=[]
                    ),
        PositionSet(date=dt.date(2024, 5, 1),
                    reference_notional=1000,
                    positions=[Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag1", value="tagvalue1")]),
                               Position(identifier='GS UN',
                                        asset_id="MA4B66MW5E27UAHKG34",
                                        name="GS",
                                        weight=-0.8,
                                        tags=[PositionTag(name="tag2", value="tagvalue2")]),
                               Position(identifier='AAPL UW',
                                        asset_id="MA4B66MW5E27U9VBB94",
                                        name="Apple",
                                        weight=0.5,
                                        tags=[PositionTag(name="tag3", value="tagvalue3")])
                               ]
                    )
    ]

    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=[same_assets_different_weights_pricing_results[1]])

    PositionSet.price_many(different_tags_common_asset_id_positions)

    for position_set in different_tags_common_asset_id_positions:
        if position_set.date == dt.date(2024, 4, 30):
            assert not position_set.positions
            assert not position_set.unpriced_positions

        else:
            for position in position_set.positions:
                if position.identifier == "GS UN":
                    assert position.asset_id == "MA4B66MW5E27UAHKG34"
                    assert position.name == "GS"
                    assert position.weight in [0.42671, -0.72]
                    assert position.quantity in [1, -1.5]

                    if position.weight == 0.42671:
                        assert position.tags[0] == PositionTag(name="tag1", value="tagvalue1")
                    else:
                        assert position.tags[0] == PositionTag(name="tag2", value="tagvalue2")
                else:
                    assert position.asset_id == "MA4B66MW5E27U9VBB94"
                    assert position.name == "Apple"
                    assert position.weight == 0.51099
                    assert position.tags[0] == PositionTag(name="tag3", value="tagvalue3")

    # No positions have tags
    position_set_list_no_tags = deepcopy(position_sets_with_tags_and_notional)

    for pos_set in position_set_list_no_tags:
        [setattr(position, 'tags', None) for position in pos_set.positions]
    mocker.patch.object(GsPriceApi, "price_many_positions",
                        return_value=expected_position_pricing_result)
    PositionSet.price_many(position_set_list_no_tags)
    for position_set in position_set_list_no_tags:
        priced_positions = position_set.positions
        for position in priced_positions:
            if position.identifier == "GS UN":
                assert position.asset_id == "MA4B66MW5E27UAHKG34"
                assert position.name == "GS"
                assert position.weight == 0.42671
                assert position.quantity == 1
                assert not position.tags
            else:
                assert position.asset_id == "MA4B66MW5E27U9VBB94"
                assert position.name == "Apple"
                assert position.weight == 0.51099
                assert position.quantity == 1
                assert not position.tags



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import asyncio
import base64
import datetime as dt
import json
import logging
import math
import os
import sys
import time
from socket import gaierror
from typing import Iterable, Optional, Union

import msgpack
from opentracing import Span, Format

from gs_quant.api.risk import RiskApi
from gs_quant.context_base import nullcontext
from gs_quant.risk import RiskRequest
from gs_quant.target.risk import OptimizationRequest
from gs_quant.tracing import Tracer

_logger = logging.getLogger(__name__)


class WebsocketUnavailable(Exception):
    pass


class GsRiskApi(RiskApi):
    USE_MSGPACK = True
    POLL_FOR_BATCH_RESULTS = False
    PRICING_API_VERSION = None

    @classmethod
    def calc_multi(cls, requests: Iterable[RiskRequest]) -> dict:
        requests = tuple(requests)
        results = cls._exec(requests)

        if len(results) < len(requests):
            results = [RuntimeError('Missing results')] * len(requests)

        return dict(zip(requests, results))

    @classmethod
    def calc(cls, request: RiskRequest) -> Iterable:
        return cls._exec(request)

    @classmethod
    def _exec(cls, request: Union[RiskRequest, Iterable[RiskRequest]]) -> Union[Iterable, dict]:
        use_msgpack = cls.USE_MSGPACK and not isinstance(request, RiskRequest)
        headers = {'Content-Type': 'application/x-msgpack'} if use_msgpack else {}
        risk_session = cls.get_session()
        version = GsRiskApi.PRICING_API_VERSION or risk_session.api_version
        result, request_id = risk_session._post(f'/{version}' + cls.__url(request),
                                                request,
                                                include_version=False,
                                                request_headers=headers,
                                                timeout=181,
                                                return_request_id=True)

        for sub_request in request:
            sub_request._id = request_id

        return result

    @classmethod
    def __url(cls, request: Union[RiskRequest, Iterable[RiskRequest]]):
        is_bulk = not isinstance(request, RiskRequest)
        return '/risk/calculate{}'.format('/bulk' if is_bulk else '')

    @classmethod
    async def get_results(cls, responses: asyncio.Queue, results: asyncio.Queue,
                          timeout: Optional[int] = None, span: Optional[Span] = None) -> Optional[str]:
        if cls.POLL_FOR_BATCH_RESULTS:
            return await cls.__get_results_poll(responses, results, timeout=timeout, span=span)
        else:
            try:
                return await cls.__get_results_ws(responses, results, timeout=timeout, span=span)
            except WebsocketUnavailable:
                return await cls.__get_results_poll(responses, results, timeout=timeout, span=span)

    @classmethod
    async def __get_results_poll(cls, responses: asyncio.Queue, results: asyncio.Queue, timeout: Optional[int] = None,
                                 span: Optional[Span] = None):
        run = True
        pending_requests = {}
        end_time = dt.datetime.now() + dt.timedelta(seconds=timeout) if timeout else None

        if span:
            Tracer.get_instance().scope_manager.activate(span, finish_on_close=False)

        while pending_requests or run:
            # Check for timeout
            if end_time is not None and dt.datetime.now() > end_time:
                _logger.error('Fatal error: timeout while waiting for results')
                cls.shutdown_queue_listener(results)
                return

            shutdown, items = await cls.drain_queue_async(responses, timeout=2)

            if shutdown:
                run = False

            if items:
                # ... update the pending requests ...
                pending_requests.update(((i[1]['reportId'], i[0]) for i in items))

            if not pending_requests:
                continue

            # ... poll for completed requests ...

            try:
                risk_session = cls.get_session()
                version = GsRiskApi.PRICING_API_VERSION or risk_session.api_version
                calc_results = risk_session._post(f'/{version}/risk/calculate/results/bulk',
                                                  list(pending_requests.keys()), include_version=False)

                # ... enqueue the request and result for the listener to handle ...
                for result in calc_results:
                    if 'error' in result:
                        results.put_nowait((pending_requests.pop(result['requestId']), RuntimeError(result['error'])))
                    elif 'result' in result:
                        results.put_nowait((pending_requests.pop(result['requestId']), result['result']))
            except Exception as e:
                error_str = f'Fatal error polling for results: {e}'
                _logger.error(error_str)
                cls.shutdown_queue_listener(results)
                return error_str

    @classmethod
    async def __get_results_ws(cls, responses: asyncio.Queue, results: asyncio.Queue, timeout: Optional[int] = None,
                               span: Optional[Span] = None):
        async def handle_websocket():
            ret = ''

            try:
                # If we're re-connecting then re-send any in-flight request ids

                outstanding_request_ids = [i for i in pending_requests.keys() if i not in dispatched]
                if outstanding_request_ids:
                    _logger.info(f'Re-sending {len(outstanding_request_ids)} requests')
                    await asyncio.wait_for(ws.send(json.dumps(outstanding_request_ids)), timeout=send_timeout)

                all_requests_dispatched = False

                while pending_requests or not all_requests_dispatched:
                    # Continue while we have pending or un-dispatched requests
                    _logger.debug(f'waiting for {", ".join(pending_requests.keys())}')
                    request_listener = asyncio.ensure_future(cls.drain_queue_async(responses)) \
                        if not all_requests_dispatched else None
                    result_listener = asyncio.ensure_future(ws.recv())
                    listeners = tuple(filter(None, (request_listener, result_listener)))

                    # Wait for either a request or result
                    complete, pending = await asyncio.wait(listeners, return_when=asyncio.FIRST_COMPLETED)

                    # Check results before sending more requests. Results can be lost otherwise
                    if result_listener in complete:
                        # New results have been received
                        request_id = None
                        try:
                            raw_res = result_listener.result()
                            parsed_res = raw_res.decode() if isinstance(raw_res, bytes) else raw_res
                            request_id, status_result_str = parsed_res.split(';', 1)
                            status, result_str = status_result_str[0], status_result_str[1:]
                        except Exception as ee:
                            status = 'E'
                            result_str = str(ee)

                        if status == 'E':
                            # An error
                            result = RuntimeError(result_str)
                        else:
                            # Unpack the result
                            try:
                                result = msgpack.unpackb(base64.b64decode(result_str), raw=False) \
                                    if cls.USE_MSGPACK else json.loads(result_str)
                            except Exception as ee:
                                result = ee
                        if request_id is None:
                            # Certain fatal websocket errors (e.g. ConnectionClosed) that are caught above will mean
                            # we have no request_id - In this case we abort and set the error on all results
                            result_listener.cancel()
                            for req in pending_requests.values():
                                results.put_nowait((req, result))
                            # Give up
                            pending_requests.clear()
                            all_requests_dispatched = True
                        else:
                            # Enqueue the request and result for the listener to handle
                            results.put_nowait((pending_requests.pop(request_id), result))
                    else:
                        result_listener.cancel()

                    if request_listener:
                        if request_listener in complete:
                            # New requests have been posted ...

                            all_requests_dispatched, items = request_listener.result()
                            if items:
                                if not all([isinstance(i[1], dict) for i in items]):
                                    error_item = next(i[1] for i in items if not isinstance(i[1], dict))
                                    raise RuntimeError(error_item[0][0][0]['errorString'])

                                # ... extract the request IDs ...
                                request_ids = [i[1]['reportId'] for i in items]

                                # ... update the pending requests ...
                                pending_requests.update(zip(request_ids, (i[0] for i in items)))

                                # ... add to our result subscription ...
                                await asyncio.wait_for(ws.send(json.dumps(request_ids)), timeout=send_timeout)

                                # ... note dispatched
                                dispatched.update(request_ids)
                        else:
                            request_listener.cancel()
            except Exception as ee:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                ret = f'{exc_type} {fname} ln:{exc_tb.tb_lineno}' + str(ee)

            return ret

        pending_requests = {}
        dispatched = set()
        error = ''
        attempts = 0
        max_attempts = 5
        send_timeout = 30

        from websockets import ConnectionClosedError
        while attempts < max_attempts:
            if attempts > 0:
                await asyncio.sleep(math.pow(2, attempts))
                _logger.error(f'{error} error, retrying (attempt {attempts + 1} of {max_attempts})')

            try:
                risk_session = cls.get_session()
                api_version = GsRiskApi.PRICING_API_VERSION or risk_session.api_version
                ws_url = f'/{api_version}/risk/calculate/results/subscribe'
                trace = Tracer(f'wss:/{ws_url}') if span else nullcontext
                with trace as scope:
                    tracing_headers = {}
                    if scope and scope.span:
                        Tracer.inject(Format.HTTP_HEADERS, tracing_headers)
                    async with risk_session._connect_websocket(ws_url, tracing_headers, include_version=False) as ws:
                        if scope and scope.span:
                            scope.span.set_tag('wss.host', ws.request_headers.get('host'))
                        error = await handle_websocket()

                attempts = max_attempts
            except ConnectionClosedError as cce:
                error = 'Connection failed: ' + str(cce)
                attempts += 1
            except asyncio.TimeoutError:
                error = 'Timed out'
                attempts = max_attempts
            except gaierror:
                raise WebsocketUnavailable()
            except Exception as e:
                error = str(e)
                attempts = max_attempts

        if error != '':
            _logger.error(f'Fatal error with websocket: {error}')
            if span:
                span.set_tag('error', True)
                span.log_kv({'event': 'error', 'message': error})
            cls.shutdown_queue_listener(results)
            return error

    @classmethod
    def create_pretrade_execution_optimization(cls, request: OptimizationRequest) -> str:
        try:
            response = cls.get_session()._post(r'/risk/execution/pretrade', request)
            _logger.info('New optimization is created with id: {}'.format(response.get("optimizationId")))
            return response
        except Exception as e:
            error = str(e)
            _logger.error(error)
            return error

    @classmethod
    def get_pretrade_execution_optimization(cls, optimization_id: str, max_attempts: int = 15):
        url = '/risk/execution/pretrade/{}/results'.format(optimization_id)
        attempts = 0
        start = time.perf_counter()
        results = {}

        while attempts < max_attempts:
            if attempts > 0:
                time.sleep(math.pow(2, attempts))
                _logger.error('Retrying (attempt {} of {})'.format(attempts, max_attempts))
            try:
                results = cls.get_session()._get(url)
                if results.get('status') == 'Running':
                    attempts += 1
                else:
                    break
            except Exception as e:
                error = str(e)
                _logger.error(error)
                return error

        if results.get('status') == 'Running':
            _logger.info('Optimization is still running. Please retry fetching the results.')
            return results
        else:
            _logger.info('Optimization is fetched in {:.3f}s.'.format(time.perf_counter() - start))
            return results
